from enum import Enum

from odmantic import AIOEngine, Model


class TreeKind(str, Enum):
    BIG = "big"
    SMALL = "small"


class Tree(Model):
    name: str
    kind: TreeKind


sequoia = Tree(name="Sequoia", kind=TreeKind.BIG)
print(sequoia.kind)
#> TreeKind.BIG
print(sequoia.kind == "big")
#> True

spruce = Tree(name="Spruce", kind="small")
print(spruce.kind)
#> TreeKind.SMALL
print(spruce.kind == TreeKind.SMALL)
#> True

engine = AIOEngine()
await engine.save_all([sequoia, spruce])



from typing import Dict, Union

from odmantic import Model


class SimpleDictModel(Model):
    field: dict


print(SimpleDictModel(field={18: "a string", True: 42, 18.3: [1, 2, 3]}).field)
#> {18: 'a string', True: 42, 18.3: [1, 2, 3]}


class IntStrDictModel(Model):
    field: Dict[int, str]


print(IntStrDictModel(field={1: "one", 2: "two"}).field)
#> {1: 'one', 2: 'two'}


class IntBoolStrDictModel(Model):
    field: Dict[int, Union[bool, str]]


print(IntBoolStrDictModel(field={0: False, 1: True, 3: "three"}).field)
#> {0: False, 1: True, 3: 'three'}



from fastapi import FastAPI, HTTPException
from pydantic import BaseModel

from odmantic import AIOEngine, Model, ObjectId


class Tree(Model):
    name: str
    average_size: float
    discovery_year: int


app = FastAPI()

engine = AIOEngine()


@app.get("/trees/{id}", response_model=Tree)
async def get_tree_by_id(id: ObjectId):
    tree = await engine.find_one(Tree, Tree.id == id)
    if tree is None:
        raise HTTPException(404)
    return tree


class TreePatchSchema(BaseModel):
    name: str = None
    average_size: float = None
    discovery_year: float = None


@app.patch("/trees/{id}", response_model=Tree)
async def update_tree_by_id(id: ObjectId, patch: TreePatchSchema):
    tree = await engine.find_one(Tree, Tree.id == id)
    if tree is None:
        raise HTTPException(404)
    tree.model_update(patch)
    await engine.save(tree)
    return tree



from pydantic import BaseModel

from odmantic import Model, SyncEngine


class Player(Model):
    name: str
    game: str


engine = SyncEngine()


player_tlo = engine.find_one(Player, Player.name == "TLO")
print(repr(player_tlo))
#> Player(id=ObjectId(...), name='TLO', game='Starcraft')

# Create the structure of the patch object with pydantic
class PatchPlayerSchema(BaseModel):
    name: str
    game: str


# Create the patch object containing the new values
patch_object = PatchPlayerSchema(name="TheLittleOne", game="Starcraft II")
# Apply the patch to the instance
player_tlo.model_update(patch_object)

print(repr(player_tlo))
#> Player(id=ObjectId(...), name='TheLittleOne', game='Starcraft II')

# Finally persist again the new instance
engine.save(player_tlo)



from __future__ import annotations

import decimal
import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Pattern, Sequence, Tuple, Type, Union

import bson
import bson.binary
import bson.decimal128
import bson.errors
import bson.int64
import bson.regex
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic.main import BaseModel
from pydantic_core import core_schema

from odmantic.typing import Annotated, get_args, get_origin


@dataclass(frozen=True)
class WithBsonSerializer:
    """Adds a BSON serializer to use on a field when it will be saved to the database"""

    bson_serializer: Callable[[Any], Any]


def _get_bson_serializer(type_: Type[Any]) -> Callable[[Any], Any] | None:
    origin = get_origin(type_)
    if origin is not None and origin == Annotated:
        args = get_args(type_)
        for arg in args:
            if isinstance(arg, WithBsonSerializer):
                return arg.bson_serializer
    return None


class ObjectId(bson.ObjectId):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_string_or_bytes(value: Union[str, bytes]) -> bson.ObjectId:
            try:
                return bson.ObjectId(value)
            except bson.errors.InvalidId:
                raise ValueError("Invalid ObjectId")

        from_string_or_bytes_schema = core_schema.chain_schema(
            [
                core_schema.union_schema(
                    [
                        core_schema.str_schema(),
                        core_schema.bytes_schema(),
                    ]
                ),
                core_schema.no_info_plain_validator_function(
                    validate_from_string_or_bytes
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_string_or_bytes_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(bson.ObjectId),
                    from_string_or_bytes_schema,
                ],
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                str, when_used="json"
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema.str_schema())
        json_schema.update(
            examples=["5f85f36d6dfecacc68428a46", "ffffffffffffffffffffffff"],
            example="5f85f36d6dfecacc68428a46",
        )
        return json_schema


class Int64(bson.Int64):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_int(value: int) -> bson.int64.Int64:
            return bson.int64.Int64(value)

        from_int_schema = core_schema.chain_schema(
            [
                core_schema.int_schema(),
                core_schema.no_info_plain_validator_function(validate_from_int),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_int_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(bson.int64.Int64),
                    from_int_schema,
                ]
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `int`
        return handler(core_schema.int_schema())


Long = Int64


class Decimal128(bson.decimal128.Decimal128):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_decimal_repr(
            value: Union[decimal.Decimal, float, str, Tuple[int, Sequence[int], int]],
        ) -> bson.decimal128.Decimal128:
            try:
                return bson.decimal128.Decimal128(value)
            except Exception:
                raise ValueError("Invalid Decimal128 value")

        from_decimal_repr_schema = core_schema.no_info_plain_validator_function(
            validate_from_decimal_repr
        )
        return core_schema.json_or_python_schema(
            json_schema=from_decimal_repr_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(bson.decimal128.Decimal128),
                    from_decimal_repr_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: v.to_decimal(), when_used="json"
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.float_schema())


class Binary(bson.binary.Binary):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_bytes(
            value: bytes,
        ) -> bson.binary.Binary:
            return bson.binary.Binary(value)

        from_bytes_schema = core_schema.chain_schema(
            [
                core_schema.bytes_schema(),
                core_schema.no_info_plain_validator_function(validate_from_bytes),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=from_bytes_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(bson.binary.Binary),
                    from_bytes_schema,
                ]
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.bytes_schema())


def validate_pattern_from_str(
    value: str,
) -> Pattern:
    try:
        return re.compile(value)
    except Exception:
        raise ValueError("Invalid Pattern value")


def validate_regex_from_pattern(
    value: Pattern,
) -> bson.regex.Regex:
    try:
        return bson.regex.Regex(value.pattern, flags=value.flags)
    except Exception:
        raise ValueError("Invalid Regex value")


def validate_pattern_from_regex(
    value: bson.regex.Regex,
) -> Pattern:
    try:
        return re.compile(value.pattern, flags=value.flags)
    except Exception:
        raise ValueError("Invalid Pattern value")


class Regex(bson.regex.Regex):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_pattern_from_str),
                core_schema.no_info_plain_validator_function(
                    validate_regex_from_pattern
                ),
            ]
        )
        from_pattern_schema = core_schema.chain_schema(
            [
                core_schema.is_instance_schema(Pattern),
                core_schema.no_info_plain_validator_function(
                    validate_regex_from_pattern
                ),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(bson.regex.Regex),
                    from_pattern_schema,
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: v.pattern, when_used="json"
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        schema = handler(core_schema.str_schema())
        schema.update(
            examples=[r"^Foo"], example=r"^Foo", type="string", format="binary"
        )
        return schema


class __PatternPydanticAnnotation:  # cannot subclass Pattern since it's final
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        from_regex_schema = core_schema.chain_schema(
            [
                core_schema.is_instance_schema(bson.regex.Regex),
                core_schema.no_info_plain_validator_function(
                    validate_pattern_from_regex
                ),
            ]
        )

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_pattern_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(Pattern),
                    from_regex_schema,
                    from_str_schema,
                ]
            ),
        )


_Pattern = Annotated[Pattern, __PatternPydanticAnnotation]


class _datetime(datetime):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_mongo_datetime(
            d: datetime,
        ) -> datetime:
            # MongoDB does not store timezone info
            # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
            if d.tzinfo is not None and d.tzinfo.utcoffset(d) != timedelta(0):
                raise ValueError("datetime objects must be naive (no timezone info)")
            # Truncate microseconds to milliseconds to comply with Mongo behavior
            microsecs = d.microsecond - d.microsecond % 1000
            return d.replace(microsecond=microsecs)

        mongo_datetime_schema = core_schema.chain_schema(
            [
                core_schema.datetime_schema(),
                core_schema.no_info_plain_validator_function(validate_mongo_datetime),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=mongo_datetime_schema,
            python_schema=mongo_datetime_schema,
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.datetime_schema())


class _decimalDecimalPydanticAnnotation:
    """This specific BSON substitution field helps to handle the support of standard
    python Decimal objects

    https://api.mongodb.com/python/current/faq.html?highlight=decimal#how-can-i-store-decimal-decimal-instances
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_decimal128(
            value: bson.decimal128.Decimal128,
        ) -> decimal.Decimal:
            return value.to_decimal()

        decimal128_schema = core_schema.chain_schema(
            [
                core_schema.is_instance_schema(bson.decimal128.Decimal128),
                core_schema.no_info_plain_validator_function(validate_from_decimal128),
            ]
        )

        def validate_from_str(
            value: str,
        ) -> decimal.Decimal:
            try:
                return decimal.Decimal(value)
            except decimal.InvalidOperation:
                raise ValueError("Invalid decimal string")

        str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(decimal.Decimal),
                    decimal128_schema,
                    str_schema,
                ]
            ),
        )


_decimalDecimal = Annotated[
    decimal.Decimal,
    _decimalDecimalPydanticAnnotation,
    WithBsonSerializer(lambda v: bson.decimal128.Decimal128(v)),
]


BSON_TYPES_ENCODERS: Dict[Type, Callable] = {
    bson.ObjectId: str,
    bson.decimal128.Decimal128: lambda x: x.to_decimal(),  # Convert to regular decimal
    bson.regex.Regex: lambda x: x.pattern,  # TODO: document no serialization of flags
}


class BaseBSONModel(BaseModel):
    """Equivalent of `pydantic.BaseModel` supporting BSON types serialization.

    If you want to apply other custom JSON encoders, you'll need to use
    [BSON_TYPES_ENCODERS][odmantic.bson.BSON_TYPES_ENCODERS] directly.
    """

    model_config = {"json_encoders": BSON_TYPES_ENCODERS}


_BSON_SUBSTITUTED_FIELDS = {
    bson.ObjectId: ObjectId,
    bson.int64.Int64: Int64,
    bson.decimal128.Decimal128: Decimal128,
    bson.binary.Binary: Binary,
    bson.regex.Regex: Regex,
    Pattern: _Pattern,
    decimal.Decimal: _decimalDecimal,
    datetime: _datetime,
}



import dataclasses
import re
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Generic, List, Pattern, Tuple, Type, TypeVar, Union

import pytest
from bson import Binary, Decimal128, Int64, ObjectId, Regex
from motor.motor_asyncio import AsyncIOMotorDatabase
from pymongo.database import Database

from odmantic.bson import WithBsonSerializer
from odmantic.engine import AIOEngine, SyncEngine
from odmantic.model import Model
from odmantic.typing import Annotated

pytestmark = pytest.mark.asyncio

T = TypeVar("T")


@dataclasses.dataclass
class TypeTestCase(Generic[T]):
    python_type: Type[T]
    bson_type: str
    sample_value: T


MIN_INT32 = -(2**31)
UNDER_INT32_VALUE = MIN_INT32 - 1
MAX_INT32 = 2**31 - 1
OVER_INT32_VALUE = MAX_INT32 + 1

sample_datetime = datetime.now()

type_test_data = [
    # Simple types
    TypeTestCase(int, "int", 15),
    TypeTestCase(int, "int", MIN_INT32),
    TypeTestCase(int, "int", MAX_INT32),
    TypeTestCase(int, "long", UNDER_INT32_VALUE),
    TypeTestCase(int, "long", OVER_INT32_VALUE),
    TypeTestCase(Int64, "long", 13),
    TypeTestCase(Int64, "long", Int64(13)),
    TypeTestCase(str, "string", "foo"),
    TypeTestCase(float, "double", 3.14),
    TypeTestCase(Decimal, "decimal", Decimal("3.14159265359")),
    TypeTestCase(
        Decimal, "decimal", "3.14159265359"
    ),  # TODO split tests for  odmantic type inference
    TypeTestCase(Decimal128, "decimal", Decimal128(Decimal("3.14159265359"))),
    TypeTestCase(Dict[str, Any], "object", {"foo": "bar", "fizz": {"foo": "bar"}}),
    TypeTestCase(bool, "bool", False),
    TypeTestCase(Pattern, "regex", re.compile(r"^.*$")),
    TypeTestCase(Pattern, "regex", re.compile(r"^.*$", flags=re.IGNORECASE)),
    TypeTestCase(
        Pattern, "regex", re.compile(r"^.*$", flags=re.IGNORECASE | re.MULTILINE)
    ),
    TypeTestCase(Regex, "regex", Regex(r"^.*$", flags=32)),
    TypeTestCase(ObjectId, "objectId", ObjectId()),
    TypeTestCase(bytes, "binData", b"\xf0\xf1\xf2"),
    TypeTestCase(Binary, "binData", Binary(b"\xf0\xf1\xf2")),
    TypeTestCase(datetime, "date", sample_datetime),
    TypeTestCase(List[str], "array", ["one"]),
    # Compound Types
    TypeTestCase(Tuple[str, ...], "array", ("one",)),  # type: ignore
    TypeTestCase(List[ObjectId], "array", [ObjectId() for _ in range(5)]),
    TypeTestCase(
        Union[Tuple[ObjectId, ...], None],  # type: ignore
        "array",
        tuple(ObjectId() for _ in range(5)),
    ),
]


def id_from_test_case(case: TypeTestCase):
    return f"{case.bson_type}"


@pytest.mark.parametrize("case", type_test_data, ids=id_from_test_case)
async def test_bson_type_inference(
    motor_database: AsyncIOMotorDatabase, aio_engine: AIOEngine, case: TypeTestCase
):
    class ModelWithTypedField(Model):
        field: case.python_type  # type: ignore

    # TODO: Fix objectid optional (type: ignore)
    instance = await aio_engine.save(ModelWithTypedField(field=case.sample_value))
    document = await motor_database[ModelWithTypedField.__collection__].find_one(
        {
            +ModelWithTypedField.id: instance.id,  # type: ignore
            +ModelWithTypedField.field: {"$type": case.bson_type},
        }
    )
    assert document is not None, (
        f"Type inference error: {case.python_type} -> {case.bson_type}"
        f" ({case.sample_value})"
    )
    recovered_instance = ModelWithTypedField(field=document["field"])
    assert recovered_instance.field == instance.field


@pytest.mark.parametrize("case", type_test_data, ids=id_from_test_case)
def test_sync_bson_type_inference(
    pymongo_database: Database, sync_engine: SyncEngine, case: TypeTestCase
):
    class ModelWithTypedField(Model):
        field: case.python_type  # type: ignore

    # TODO: Fix objectid optional (type: ignore)
    instance = sync_engine.save(ModelWithTypedField(field=case.sample_value))
    document = pymongo_database[ModelWithTypedField.__collection__].find_one(
        {
            +ModelWithTypedField.id: instance.id,  # type: ignore
            +ModelWithTypedField.field: {"$type": case.bson_type},
        }
    )
    assert document is not None, (
        f"Type inference error: {case.python_type} -> {case.bson_type}"
        f" ({case.sample_value})"
    )
    recovered_instance = ModelWithTypedField(field=document["field"])
    assert recovered_instance.field == instance.field


async def test_custom_bson_serializable(
    motor_database: AsyncIOMotorDatabase, aio_engine
):
    FancyFloat = Annotated[float, WithBsonSerializer(str)]

    class ModelWithCustomField(Model):
        field: FancyFloat

    instance = await aio_engine.save(ModelWithCustomField(field=3.14))
    document = await motor_database[ModelWithCustomField.__collection__].find_one(
        {
            +ModelWithCustomField.id: instance.id,  # type: ignore
            +ModelWithCustomField.field: {"$type": "string"},
        }
    )
    assert document is not None, "Couldn't retrieve the document with it's string value"
    recovered_instance = ModelWithCustomField.model_validate_doc(document)
    assert recovered_instance.field == instance.field


def test_sync_custom_bson_serializable(
    pymongo_database: Database, sync_engine: SyncEngine
):
    FancyFloat = Annotated[float, WithBsonSerializer(str)]

    class ModelWithCustomField(Model):
        field: FancyFloat

    instance = sync_engine.save(ModelWithCustomField(field=3.14))
    document = pymongo_database[ModelWithCustomField.__collection__].find_one(
        {
            +ModelWithCustomField.id: instance.id,  # type: ignore
            +ModelWithCustomField.field: {"$type": "string"},
        }
    )
    assert document is not None, "Couldn't retrieve the document with it's string value"
    recovered_instance = ModelWithCustomField.model_validate_doc(document)
    assert recovered_instance.field == instance.field



from typing import Annotated
from odmantic import AIOEngine, Model, WithBsonSerializer

class ASCIISerializedAsBinaryBase(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, bytes):  # Handle data coming from MongoDB
            return v.decode("ascii")
        if not isinstance(v, str):
            raise TypeError("string required")
        if not v.isascii():
            raise ValueError("Only ascii characters are allowed")
        return v


def serialize_ascii_to_bytes(v: ASCIISerializedAsBinaryBase) -> bytes:
    # We can encode this string as ascii since it contains
    # only ascii characters
    bytes_ = v.encode("ascii")
    return bytes_


ASCIISerializedAsBinary = Annotated[
    ASCIISerializedAsBinaryBase, WithBsonSerializer(serialize_ascii_to_bytes)
]

class Example(Model):
    field: ASCIISerializedAsBinary

engine = AIOEngine()
await engine.save(Example(field="hello world"))
fetched = await engine.find_one(Example)
print(fetched.field)
#> hello world



"""Models based on https://github.com/tortoise/orm-benchmarks"""

from datetime import datetime, timezone
from decimal import Decimal
from random import choice
from typing import Iterator, Optional

from typing_extensions import Literal, get_args

from odmantic import Field, Model

Level = Literal[10, 20, 30, 40, 50]
VALID_LEVELS = list(get_args(Level))


def utc_now():
    return datetime.now(timezone.utc)


class SmallJournal(Model):
    timestamp: datetime = Field(default_factory=utc_now)
    level: Level = Field(index=True)
    text: str = Field(index=True)

    @classmethod
    def get_random_instances(cls, context: str, count: int) -> Iterator["SmallJournal"]:
        for i in range(count):
            yield cls(level=choice(VALID_LEVELS), text=f"From {context}, item {i}")


class JournalWithRelations(Model):
    timestamp: datetime = Field(default_factory=utc_now)
    level: Level = Field(index=True)
    text: str = Field(index=True)

    # parent


class BigJournal(Model):
    timestamp: datetime = Field(default_factory=utc_now)
    level: Level = Field(index=True)
    text: str = Field(index=True)

    col_float1: float = Field(default=2.2)
    col_smallint1: int = Field(default=2)
    col_int1: int = Field(default=2000000)
    col_bigint1: int = Field(default=99999999)
    col_char1: str = Field(default=255, max_length=255)
    col_text1: str = Field(
        default="Moo,Foo,Baa,Waa,Moo,Foo,Baa,Waa,Moo,Foo,Baa,Waa",
    )
    col_decimal1: Decimal = Field(default=Decimal("2.2"))
    col_json1: dict = Field(
        default={"a": 1, "b": "b", "c": [2], "d": {"e": 3}, "f": True},
    )

    col_float2: Optional[float] = Field(default=None)
    col_smallint2: Optional[int] = Field(default=None)
    col_int2: Optional[int] = Field(default=None)
    col_bigint2: Optional[int] = Field(default=None)
    col_char2: Optional[str] = Field(default=None, max_length=255)
    col_text2: Optional[str] = Field(
        default=None,
    )
    col_decimal2: Optional[Decimal] = Field(default=None)
    col_json2: Optional[dict] = Field(
        default=None,
    )



from odmantic import Field, Model, SyncEngine


class Player(Model):
    name: str = Field(unique=True)


engine = SyncEngine()
engine.configure_database([Player])

leeroy = Player(name="Leeroy")
engine.save(leeroy)

another_leeroy = Player(name="Leeroy")
engine.save(another_leeroy)
#> Raises odmantic.exceptions.DuplicateKeyError:
#>    Duplicate key error for: Player.
#>    Instance: id=ObjectId('6314b4c25a19444bfe0c0be5') name='Leeroy'



from typing import ClassVar

from pydantic import ValidationError, validator

from odmantic import Model


class SmallRectangle(Model):
    MAX_SIDE_SIZE: ClassVar[float] = 10

    length: float
    width: float

    @validator("width", "length")
    def check_small_sides(cls, v):
        if v > cls.MAX_SIDE_SIZE:
            raise ValueError(f"side is greater than {cls.MAX_SIDE_SIZE}")
        return v

    @validator("width")
    def check_width_length(cls, width, values, **kwargs):
        length = values.get("length")
        if length is not None and width > length:
            raise ValueError("width can't be greater than length")
        return width


print(SmallRectangle(length=2, width=1))
#> id=ObjectId('5f81e3c073103f509f97e374'), length=2.0, width=1.0

try:
    SmallRectangle(length=2)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
    width
      field required (type=value_error.missing)
    """

try:
    SmallRectangle(length=2, width=3)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
    width
      width can't be greater than length (type=value_error)
    """

try:
    SmallRectangle(length=40, width=3)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
    length
      side is greater than 10 (type=value_error)
    """



from typing import List

from odmantic import AIOEngine, EmbeddedModel, Model


class Address(EmbeddedModel):
    street: str
    city: str
    state: str
    zipcode: str


class Customer(Model):
    name: str
    addresses: List[Address]


customer = Customer(
    name="John Doe",
    addresses=[
        Address(
            street="1757  Birch Street",
            city="Greenwood",
            state="Indiana",
            zipcode="46142",
        ),
        Address(
            street="262  Barnes Avenue",
            city="Cincinnati",
            state="Ohio",
            zipcode="45216",
        ),
    ],
)

engine = AIOEngine()
await engine.save(customer)



from odmantic import AIOEngine, Model


class Player(Model):
    name: str
    game: str


engine = AIOEngine()

async with engine.transaction() as transaction:
    await transaction.save(Player(name="Leeroy Jenkins", game="WoW"))
    await transaction.commit()

print(engine.count(Player))
#> 1

async with engine.transaction() as transaction:
    await transaction.save(Player(name="Shroud", game="Counter-Strike"))
    await transaction.save(Player(name="Serral", game="Starcraft"))
    await transaction.abort()

print(engine.count(Player))
#> 1



import json
from datetime import datetime
from typing import Dict

import pytest
from bson import ObjectId
from bson.binary import Binary
from bson.decimal128 import Decimal128
from bson.int64 import Int64
from bson.regex import Regex

from odmantic import Model
from tests.zoo.book_embedded import Book, Publisher
from tests.zoo.full_bson import FullBsonModel
from tests.zoo.patron_embedded import Address, Patron
from tests.zoo.person import PersonModel
from tests.zoo.tree import TreeKind, TreeModel
from tests.zoo.twitter_user import TwitterUser

pytestmark = pytest.mark.asyncio


def test_simple_model_serialization():
    class M(Model): ...

    id_ = ObjectId()
    assert json.loads(M(id=id_).model_dump_json()) == {"id": str(id_)}


TWITTER_USERS = [TwitterUser(), TwitterUser(), TwitterUser()]
MAIN_TWITTER_USER = TwitterUser(following=[e.id for e in TWITTER_USERS])


@pytest.mark.parametrize(
    "instance, expected_parsed_json",
    (
        (
            PersonModel(first_name="Johnny", last_name="Cash"),
            dict(first_name="Johnny", last_name="Cash"),
        ),
        (
            TreeModel(
                name="Secoya",
                average_size=100.3,
                discovery_year=1253,
                kind=TreeKind.BIG,
                genesis_continents=["Asia"],
                per_continent_density={"Asia": 20.3},
            ),
            dict(
                name="Secoya",
                average_size=100.3,
                discovery_year=1253,
                kind="big",
                genesis_continents=["Asia"],
                per_continent_density={"Asia": 20.3},
            ),
        ),
        (
            Book(
                title="Harry Potter",
                pages=550,
                publisher=Publisher(name="A publisher", founded=1995, location="CA"),
            ),
            dict(
                title="Harry Potter",
                pages=550,
                publisher=dict(name="A publisher", founded=1995, location="CA"),
            ),
        ),
        (
            Patron(
                name="Jean Michel",
                addresses=[
                    Address(
                        street="212 Rue de Tolbiac",
                        city="Paris",
                        state="Ile de France",
                        zip="75013",
                    )
                ],
            ),
            dict(
                name="Jean Michel",
                addresses=[
                    dict(
                        street="212 Rue de Tolbiac",
                        city="Paris",
                        state="Ile de France",
                        zip="75013",
                    )
                ],
            ),
        ),
        (MAIN_TWITTER_USER, dict(following=[str(u.id) for u in TWITTER_USERS])),
        (
            FullBsonModel(
                objectId_=ObjectId("5f6bd0f85cac5a450e8eb9e8"),
                long_=Int64(258),
                decimal_=Decimal128("256.123457"),
                # TODO: document some bytes value might be rejected because of utf8
                # encoding: encode in base64 before
                binary_=Binary(b"\x48\x49"),
                regex_=Regex(r"^.*$"),
            ),
            dict(
                objectId_="5f6bd0f85cac5a450e8eb9e8",
                long_=258,
                decimal_="256.123457",
                binary_=b"\x48\x49".decode(),
                regex_=r"^.*$",
            ),
        ),
    ),
)
def test_zoo_serialization_no_id(instance: Model, expected_parsed_json: Dict):
    parsed_data = json.loads(instance.model_dump_json())
    del parsed_data["id"]
    assert parsed_data == expected_parsed_json


@pytest.mark.filterwarnings("ignore:`json_encoders` is deprecated")
def test_custom_json_encoders():
    class M(Model):
        a: datetime = datetime.now()

        model_config = {"json_encoders": {datetime: lambda _: "encoded"}}

    instance = M()
    parsed = json.loads(instance.model_dump_json())
    assert parsed == {"id": str(instance.id), "a": "encoded"}


@pytest.mark.xfail(
    reason=(
        "This doesn't work any more with pydantic v2 since bson fields are "
        "now annotated and take precedence over the custom json encoder."
    )
)
def test_custom_json_encoders_override_builtin_bson():
    class M(Model):
        model_config = {"json_encoders": {ObjectId: lambda _: "encoded"}}

    instance = M()
    parsed = json.loads(instance.model_dump_json())
    assert parsed == {"id": "encoded"}



await engine.find_one(
    Country, Country.capital_city.name == "Stockholm"
)
#> Country(
#>     id=ObjectId("5f79d7e8b305f24ca43593e2"),
#>     name="Sweden",
#>     currency="Swedish krona",
#>     capital_city=CapitalCity(name="Stockholm", population=975904),
#> )



from odmantic import Model, SyncEngine


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

player_tlo = engine.find_one(Player, Player.name == "TLO")
print(repr(player_tlo))
#> Player(id=ObjectId(...), name='TLO', game='Starcraft')

# Create the patch dictionary containing the new values
patch_object = {"name": "TheLittleOne", "game": "Starcraft II"}
# Update the local instance
player_tlo.model_update(patch_object)

print(repr(player_tlo))
#> Player(id=ObjectId(...), name='TheLittleOne', game='Starcraft II')

# Finally persist the instance
engine.save(player_tlo)



import re


def is_dunder(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def raise_on_invalid_key_name(name: str) -> None:
    # https://docs.mongodb.com/manual/reference/limits/#Restrictions-on-Field-Names
    if name.startswith("$"):
        raise TypeError("key_name cannot start with the dollar sign ($) character")
    if "." in name:
        raise TypeError("key_name cannot contain the dot (.) character")


def raise_on_invalid_collection_name(collection_name: str, cls_name: str) -> None:
    # https://docs.mongodb.com/manual/reference/limits/#Restriction-on-Collection-Names
    if "$" in collection_name:
        raise TypeError(f"Invalid collection name for {cls_name}: cannot contain '$'")
    if collection_name == "":
        raise TypeError(f"Invalid collection name for {cls_name}: cannot be empty")
    if collection_name.startswith("system."):
        raise TypeError(
            f"Invalid collection name for {cls_name}:" " cannot start with 'system.'"
        )


def to_snake_case(s: str) -> str:
    tmp = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", tmp).lower()


class Undefined:
    pass



from datetime import datetime

from odmantic import Index, Model
from odmantic.query import asc, desc


class Event(Model):
    username: str
    date: datetime

    model_config = {"indexes": lambda: [Index(asc(Event.username), desc(Event.date))]}



def define_env(env):
    @env.macro
    def async_sync_snippet(folder: str, filename: str, hl_lines=None, linenums=True):
        return f"""
=== "Async"
    ```python {'linenums="1"' if linenums else ''} {'hl_lines="'+hl_lines+'"' if hl_lines is not None else ''}"
    --8<-- "{folder}/async/{filename}"
    ```

=== "Sync"
    ```python {'linenums="1"' if linenums else ''} {'hl_lines="'+hl_lines+'"' if hl_lines is not None else ''}"
    --8<-- "{folder}/sync/{filename}"
    ```
"""



from odmantic import AIOEngine, Model, Reference


class Publisher(Model):
    name: str
    founded: int
    location: str


class Book(Model):
    title: str
    pages: int
    publisher: Publisher = Reference()


hachette = Publisher(name="Hachette Livre", founded=1826, location="FR")
harper = Publisher(name="HarperCollins", founded=1989, location="US")

books = [
    Book(title="They Didn't See Us Coming", pages=304, publisher=hachette),
    Book(title="This Isn't Happening", pages=256, publisher=hachette),
    Book(title="Prodigal Summer", pages=464, publisher=harper),
]

engine = AIOEngine()
await engine.save_all(books)



from datetime import datetime

from bson import ObjectId

from odmantic import Model
from odmantic.exceptions import DocumentParsingError
from odmantic.field import Field


class User(Model):
    name: str
    updated_at: datetime = Field(default_factory=datetime.utcnow)

    model_config = {"parse_doc_with_default_factories": True}


document = {"name": "Leeroy", "_id": ObjectId("5f8352a87a733b8b18b0cb27")}

user = User.model_validate_doc(document)
print(repr(user))
#> User(
#>     id=ObjectId("5f8352a87a733b8b18b0cb27"),
#>     name="Leeroy",
#>     updated_at=datetime.datetime(2020, 11, 8, 23, 28, 19, 980000),
#> )



from typing import List

from odmantic.model import EmbeddedModel, Model


class Address(EmbeddedModel):
    street: str
    city: str
    state: str
    zip: str


class Patron(Model):
    name: str
    addresses: List[Address]



from odmantic import AIOEngine, Model


class Player(Model):
    name: str
    game: str


engine = AIOEngine()

players = await engine.find(Player)

for player in players:
    await engine.delete(player)



import uvicorn
from fastapi import FastAPI, HTTPException

from odmantic import AIOEngine, Model, ObjectId


class Tree(Model):
    name: str
    average_size: float
    discovery_year: int


app = FastAPI()

engine = AIOEngine()


@app.get("/trees/{id}", response_model=Tree)
async def get_tree_by_id(id: ObjectId):
    tree = await engine.find_one(Tree, Tree.id == id)
    if tree is None:
        raise HTTPException(404)
    return tree


@app.delete("/trees/{id}", response_model=Tree)
async def delete_tree_by_id(id: ObjectId):
    tree = await engine.find_one(Tree, Tree.id == id)
    if tree is None:
        raise HTTPException(404)
    await engine.delete(tree)
    return tree



from odmantic import SyncEngine, Model


class User(Model):
    name: str


engine = SyncEngine()
collection = engine.get_collection(User)
print(collection)
#> Collection(
#>     Database(
#>         MongoClient(
#>             host=["localhost:27017"],
#>             document_class=dict,
#>             tz_aware=False,
#>             connect=True,
#>         ),
#>         "test",
#>     ),
#>     "user",
#> )



from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import pymongo
from pymongo import MongoClient
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.command_cursor import CommandCursor
from pymongo.database import Database

from odmantic.exceptions import DocumentNotFoundError, DuplicateKeyError
from odmantic.field import FieldProxy, ODMReference
from odmantic.index import ODMBaseIndex
from odmantic.model import Model
from odmantic.query import QueryExpression, SortExpression, and_
from odmantic.session import (
    AIOSession,
    AIOSessionBase,
    AIOTransaction,
    SyncSession,
    SyncSessionBase,
    SyncTransaction,
)
from odmantic.typing import lenient_issubclass

try:
    import motor
    from motor.motor_asyncio import (
        AsyncIOMotorClient,
        AsyncIOMotorClientSession,
        AsyncIOMotorCollection,
        AsyncIOMotorCursor,
        AsyncIOMotorDatabase,
    )
except ImportError:  # pragma: no cover
    motor = None


ModelType = TypeVar("ModelType", bound=Model)

SortExpressionType = Optional[Union[FieldProxy, Tuple[FieldProxy]]]

AIOSessionType = Union[AsyncIOMotorClientSession, AIOSession, AIOTransaction, None]
SyncSessionType = Union[ClientSession, SyncSession, SyncTransaction, None]


class BaseCursor(Generic[ModelType]):
    """This object has to be built from the [odmantic.engine.AIOEngine.find][] method.

    An AIOCursor object support multiple async operations:

      - **async for**: asynchronously iterate over the query results
      - **await** : when awaited it will return a list of the fetched models
    """

    def __init__(
        self,
        model: Type[ModelType],
        cursor: Union["AsyncIOMotorCursor", "CommandCursor"],
    ):
        self._model = model
        self._cursor = cursor
        self._results: Optional[List[ModelType]] = None

    def _parse_document(self, raw_doc: Dict) -> ModelType:
        instance = self._model.model_validate_doc(raw_doc)
        object.__setattr__(instance, "__fields_modified__", set())
        return instance


class AIOCursor(
    BaseCursor[ModelType], AsyncIterable[ModelType], Awaitable[List[ModelType]]
):
    """This object has to be built from the [odmantic.engine.AIOEngine.find][] method.

    An AIOCursor object support multiple async operations:

      - **async for**: asynchronously iterate over the query results
      - **await** : when awaited it will return a list of the fetched models
    """

    _cursor: "AsyncIOMotorCursor"

    def __init__(self, model: Type[ModelType], cursor: "AsyncIOMotorCursor"):
        super().__init__(model=model, cursor=cursor)

    def __await__(self) -> Generator[None, None, List[ModelType]]:
        if self._results is not None:
            return self._results
        raw_docs = yield from self._cursor.to_list(length=None).__await__()
        instances = []
        for raw_doc in raw_docs:
            instances.append(self._parse_document(raw_doc))
            yield
        self._results = instances
        return instances

    async def __aiter__(self) -> AsyncGenerator[ModelType, None]:
        if self._results is not None:
            for res in self._results:
                yield res
            return
        results = []
        async for raw_doc in self._cursor:
            instance = self._parse_document(raw_doc)
            results.append(instance)
            yield instance
        self._results = results


class SyncCursor(BaseCursor[ModelType], Iterable[ModelType]):
    """This object has to be built from the [odmantic.engine.SyncEngine.find][] method.

    A SyncCursor object supports iterating over the query results using **`for`**.

    To get a list of all the results you can wrap it with `list`, as in `list(cursor)`.
    """

    _cursor: "CommandCursor"

    def __init__(self, model: Type[ModelType], cursor: "CommandCursor"):
        super().__init__(model=model, cursor=cursor)

    def __iter__(self) -> Iterator[ModelType]:
        if self._results is not None:
            for res in self._results:
                yield res
            return
        results = []
        for raw_doc in self._cursor:
            instance = self._parse_document(raw_doc)
            results.append(instance)
            yield instance
        self._results = results


_FORBIDDEN_DATABASE_CHARACTERS = set(("/", "\\", ".", '"', "$"))


class BaseEngine:
    """The BaseEngine is the base class for the async and sync engines. It holds the
    common functionality, like generating the MongoDB queries, that is then used by the
    two engines.
    """

    def __init__(
        self,
        client: Union["AsyncIOMotorClient", "MongoClient"],
        database: str = "test",
    ):
        # https://docs.mongodb.com/manual/reference/limits/#naming-restrictions
        forbidden_characters = _FORBIDDEN_DATABASE_CHARACTERS.intersection(
            set(database)
        )
        if len(forbidden_characters) > 0:
            raise ValueError(
                f"database name cannot contain: {' '.join(forbidden_characters)}"
            )
        self.client = client
        self.database_name = database
        self.database = client[self.database_name]

    @staticmethod
    def _build_query(*queries: Union[QueryExpression, Dict, bool]) -> QueryExpression:
        if len(queries) == 0:
            return QueryExpression()
        for query in queries:
            if isinstance(query, bool):
                raise TypeError("cannot build query using booleans")
        queries = cast(Tuple[Union[QueryExpression, Dict], ...], queries)
        if len(queries) == 1:
            return QueryExpression(queries[0])
        return and_(*queries)

    @staticmethod
    def _cascade_find_pipeline(
        model: Type[ModelType], doc_namespace: str = ""
    ) -> List[Dict]:
        """Recursively build the find pipeline for model."""
        pipeline: List[Dict] = []
        for ref_field_name in model.__references__:
            odm_reference = cast(ODMReference, model.__odm_fields__[ref_field_name])
            pipeline.extend(
                [
                    {
                        "$lookup": {
                            "from": odm_reference.model.__collection__,
                            "let": {"foreign_id": f"${odm_reference.key_name}"},
                            "pipeline": [
                                {
                                    "$match": {
                                        "$expr": {"$eq": ["$_id", "$$foreign_id"]}
                                    }
                                },
                                *BaseEngine._cascade_find_pipeline(
                                    odm_reference.model,
                                    doc_namespace=f"{doc_namespace}{ref_field_name}.",
                                ),
                            ],
                            "as": odm_reference.key_name,
                            # FIXME if ref field name is an existing key_name ?
                        }
                    },
                    {  # Preserves document with unbound references
                        "$unwind": {
                            "path": f"${odm_reference.key_name}",
                            "preserveNullAndEmptyArrays": True,
                        }
                    },
                ]
            )
        return pipeline

    @staticmethod
    def _build_sort_expression(
        sort_field: Union[FieldProxy, SortExpression],
    ) -> SortExpression:
        return (
            SortExpression({+sort_field: 1})
            if not isinstance(sort_field, SortExpression)
            else sort_field
        )

    @classmethod
    def _validate_sort_argument(cls, sort: Any) -> Optional[SortExpression]:
        if sort is None:
            return None

        if isinstance(sort, tuple):
            for sorted_field in sort:
                if not isinstance(sorted_field, (FieldProxy, SortExpression)):
                    raise TypeError(
                        "sort elements have to be Model fields or asc, desc descriptors"
                    )
            sort_expression: Dict = {}
            for sort_field in sort:
                sort_expression.update(cls._build_sort_expression(sort_field))

            return SortExpression(sort_expression)

        if not isinstance(sort, (FieldProxy, SortExpression)):
            raise TypeError(
                "sort has to be a Model field or "
                "asc, desc descriptors or a tuple of these"
            )

        return cls._build_sort_expression(sort)

    def _prepare_find_pipeline(
        self,
        model: Type[ModelType],
        *queries: Union[
            QueryExpression, Dict, bool
        ],  # bool: allow using binary operators with mypy
        sort: Optional[Any] = None,
        skip: int = 0,
        limit: Optional[int] = None,
    ) -> List[Dict[str, Any]]:
        if not lenient_issubclass(model, Model):
            raise TypeError("Can only call find with a Model class")
        sort_expression = self._validate_sort_argument(sort)
        if limit is not None and limit <= 0:
            raise ValueError("limit has to be a strict positive value or None")
        if skip < 0:
            raise ValueError("skip has to be a positive integer")
        query = BaseEngine._build_query(*queries)
        pipeline: List[Dict[str, Any]] = [{"$match": query}]
        if sort_expression is not None:
            pipeline.append({"$sort": sort_expression})
        if skip > 0:
            pipeline.append({"$skip": skip})
        if limit is not None and limit > 0:
            pipeline.append({"$limit": limit})
        pipeline.extend(BaseEngine._cascade_find_pipeline(model))
        return pipeline


class AIOEngine(BaseEngine):
    """The AIOEngine object is responsible for handling database operations with MongoDB
    in an asynchronous way using motor.
    """

    client: "AsyncIOMotorClient"
    database: "AsyncIOMotorDatabase"

    def __init__(
        self,
        client: Union["AsyncIOMotorClient", None] = None,
        database: str = "test",
    ):
        """Engine constructor.

        Args:
            client: instance of an AsyncIO motor client. If None, a default one
                    will be created
            database: name of the database to use

        <!---
        #noqa: DAR401 RuntimeError
        -->
        """
        if not motor:
            raise RuntimeError(
                "motor is required to use AIOEngine, install it with:\n\n"
                + 'pip install "odmantic[motor]"'
            )
        if client is None:
            client = AsyncIOMotorClient()
        super().__init__(client=client, database=database)

    def get_collection(self, model: Type[ModelType]) -> "AsyncIOMotorCollection":
        """Get the motor collection associated to a Model.

        Args:
            model: model class

        Returns:
            the AsyncIO motor collection object
        """
        return self.database[model.__collection__]

    @staticmethod
    def _get_session(
        session: Union[AIOSessionType, AIOSessionBase],
    ) -> Optional[AsyncIOMotorClientSession]:
        if isinstance(session, (AIOSession, AIOTransaction)):
            return session.get_driver_session()
        assert not isinstance(session, AIOSessionBase)  # Abstract class
        return session

    async def configure_database(
        self,
        models: Sequence[Type[ModelType]],
        *,
        update_existing_indexes: bool = False,
        session: SyncSessionType = None,
    ) -> None:
        """Apply model constraints to the database.

        Args:
            models: list of models to initialize the database with
            update_existing_indexes: conflicting indexes will be dropped before creation
            session: an optional session to use for the operation

        <!---
        #noqa: DAR401 pymongo.errors.OperationFailure
        -->
        """
        driver_session = self._get_session(session)
        for model in models:
            collection = self.get_collection(model)
            for index in model.__indexes__():
                pymongo_index = (
                    index.get_pymongo_index()
                    if isinstance(index, ODMBaseIndex)
                    else index
                )
                try:
                    await collection.create_indexes(
                        [pymongo_index], session=driver_session
                    )
                except pymongo.errors.OperationFailure as exc:
                    if update_existing_indexes and getattr(exc, "code", None) in (
                        85,  # aka IndexOptionsConflict
                        86,  # aka IndexKeySpecsConflict for MongoDB > 5
                    ):
                        await collection.drop_index(
                            pymongo_index.document["name"], session=driver_session
                        )
                        await collection.create_indexes(
                            [pymongo_index], session=driver_session
                        )
                    else:
                        raise

    def session(self) -> AIOSession:
        """Get a new session for the engine to allow ordering sequential operations.

        Returns:
            a new session object

        Example usage:

        ```python
        engine = AIOEngine(...)
        async with engine.session() as session:
            john = await session.find(User, User.name == "John")
            john.name = "Doe"
            await session.save(john)
        ```
        """
        return AIOSession(self)

    def transaction(self) -> AIOTransaction:
        """Get a new transaction for the engine to aggregate sequential operations.

        Returns:
            a new transaction object

        Example usage:
        ```python
        engine = AIOEngine(...)
        async with engine.transaction() as transaction:
            john = transaction.find(User, User.name == "John")
            john.name = "Doe"
            await transaction.save(john)
            await transaction.commit()
        ```

        Warning:
            MongoDB transaction are only supported on replicated clusters: either
            directly a replicaSet or a sharded cluster with replication enabled.
        """
        return AIOTransaction(self)

    def find(
        self,
        model: Type[ModelType],
        *queries: Union[
            QueryExpression, Dict, bool
        ],  # bool: allow using binary operators with mypy
        sort: Optional[Any] = None,
        skip: int = 0,
        limit: Optional[int] = None,
        session: AIOSessionType = None,
    ) -> AIOCursor[ModelType]:
        """Search for Model instances matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            sort: sort expression
            skip: number of document to skip
            limit: maximum number of instance fetched
            session: an optional session to use for the operation

        Raises:
            DocumentParsingError: unable to parse one of the resulting documents

        Returns:
            [odmantic.engine.AIOCursor][] of the query

        <!---
        #noqa: DAR401 ValueError
        #noqa: DAR401 TypeError
        #noqa: DAR402 DocumentParsingError
        -->
        """
        pipeline = self._prepare_find_pipeline(
            model,
            *queries,
            sort=sort,
            skip=skip,
            limit=limit,
        )
        collection = self.get_collection(model)
        motor_cursor = collection.aggregate(
            pipeline, session=self._get_session(session)
        )
        return AIOCursor(model, motor_cursor)

    async def find_one(
        self,
        model: Type[ModelType],
        *queries: Union[
            QueryExpression, Dict, bool
        ],  # bool: allow using binary operators w/o plugin
        sort: Optional[Any] = None,
        session: AIOSessionType = None,
    ) -> Optional[ModelType]:
        """Search for a Model instance matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            sort: sort expression
            session: an optional session to use for the operation

        Raises:
            DocumentParsingError: unable to parse the resulting document

        Returns:
            the fetched instance if found otherwise None

        <!---
        #noqa: DAR401 TypeError
        #noqa: DAR402 DocumentParsingError
        -->
        """
        if not lenient_issubclass(model, Model):
            raise TypeError("Can only call find_one with a Model class")
        results = await self.find(model, *queries, sort=sort, limit=1, session=session)
        if len(results) == 0:
            return None
        return results[0]

    async def _save(
        self, instance: ModelType, session: "AsyncIOMotorClientSession"
    ) -> ModelType:
        """Perform an atomic save operation in the specified session"""
        for ref_field_name in instance.__references__:
            sub_instance = cast(Model, getattr(instance, ref_field_name))
            await self._save(sub_instance, session)

        fields_to_update = instance.__fields_modified__ | instance.__mutable_fields__
        if len(fields_to_update) > 0:
            doc = instance.model_dump_doc(include=fields_to_update)
            collection = self.get_collection(type(instance))
            try:
                await collection.update_one(
                    instance.model_dump_doc(include={instance.__primary_field__}),
                    {"$set": doc},
                    upsert=True,
                    session=session,
                )
            except pymongo.errors.DuplicateKeyError as e:
                raise DuplicateKeyError(instance, e)
            object.__setattr__(instance, "__fields_modified__", set())
        return instance

    async def save(
        self,
        instance: ModelType,
        *,
        session: AIOSessionType = None,
    ) -> ModelType:
        """Persist an instance to the database

        This method behaves as an 'upsert' operation. If a document already exists
        with the same primary key, it will be overwritten.

        All the other models referenced by this instance will be saved as well.

        Args:
            instance: instance to persist
            session: An optional session to use for the operation. If not provided, an
                     internal session will be used to persist the instance and
                     sub-instances.

        Returns:
            the saved instance

        Raises:
            DuplicateKeyError: the instance is duplicated according to a unique index.

        NOTE:
            The save operation actually modify the instance argument in place. However,
            the instance is still returned for convenience.

        <!---
        #noqa: DAR401 TypeError
        #noqa: DAR402 DuplicateKeyError
        -->
        """
        if not isinstance(instance, Model):
            raise TypeError("Can only call find_one with a Model class")
        if session:
            await self._save(instance, self._get_session(session))
        else:
            async with await self.client.start_session() as local_session:
                await self._save(instance, local_session)
        return instance

    async def save_all(
        self,
        instances: Sequence[ModelType],
        *,
        session: AIOSessionType = None,
    ) -> List[ModelType]:
        """Persist instances to the database

        This method behaves as multiple 'upsert' operations. If one of the document
        already exists with the same primary key, it will be overwritten.

        All the other models referenced by this instance will be recursively saved as
        well.

        Args:
            instances: instances to persist
            session: An optional session to use for the operation. If not provided, an
                     internal session will be used to persist the instances.

        Returns:
            the saved instances

        Raises:
            DuplicateKeyError: an instance is duplicated according to a unique index.

        NOTE:
            The save_all operation actually modify the arguments in place. However, the
            instances are still returned for convenience.

        <!---
        #noqa: DAR402 DuplicateKeyError
        -->
        """
        if session:
            added_instances = [
                await self._save(instance, self._get_session(session))
                for instance in instances
            ]
        else:
            async with await self.client.start_session() as local_session:
                added_instances = [
                    await self._save(instance, local_session) for instance in instances
                ]
        return added_instances

    async def delete(
        self,
        instance: ModelType,
        *,
        session: AIOSessionType = None,
    ) -> None:
        """Delete an instance from the database

        Args:
            instance: the instance to delete
            session: an optional session to use for the operation


        Raises:
            DocumentNotFoundError: the instance has not been persisted to the database
        """
        # TODO handle cascade deletion
        collection = self.database[instance.__collection__]
        pk_name = instance.__primary_field__
        result = await collection.delete_many(
            {"_id": getattr(instance, pk_name)}, session=self._get_session(session)
        )
        count = int(result.deleted_count)
        if count == 0:
            raise DocumentNotFoundError(instance)

    async def remove(
        self,
        model: Type[ModelType],
        *queries: Union[QueryExpression, Dict, bool],
        just_one: bool = False,
        session: AIOSessionType = None,
    ) -> int:
        """Delete Model instances matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            just_one: limit the deletion to just one document
            session: an optional session to use for the operation

        Returns:
            the number of instances deleted from the database.

        """
        query = AIOEngine._build_query(*queries)
        collection = self.get_collection(model)

        if just_one:
            result = await collection.delete_one(
                query, session=self._get_session(session)
            )
        else:
            result = await collection.delete_many(
                query, session=self._get_session(session)
            )

        return cast(int, result.deleted_count)

    async def count(
        self,
        model: Type[ModelType],
        *queries: Union[QueryExpression, Dict, bool],
        session: AIOSessionType = None,
    ) -> int:
        """Get the count of documents matching a query

        Args:
            model: model to perform the operation on
            *queries: query filters to apply
            session: an optional session to use for the operation

        Returns:
            number of document matching the query

        <!---
        #noqa: DAR401 TypeError
        -->
        """
        if not lenient_issubclass(model, Model):
            raise TypeError("Can only call count with a Model class")
        query = BaseEngine._build_query(*queries)
        collection = self.database[model.__collection__]
        count = await collection.count_documents(
            query, session=self._get_session(session)
        )
        return int(count)


class SyncEngine(BaseEngine):
    """The SyncEngine object is responsible for handling database operations with
    MongoDB in an synchronous way using pymongo.
    """

    client: "MongoClient"
    database: "Database"

    def __init__(
        self,
        client: "Union[MongoClient, None]" = None,
        database: str = "test",
    ):
        """Engine constructor.

        Args:
            client: instance of a PyMongo client. If None, a default one
                    will be created
            database: name of the database to use
        """
        if client is None:
            client = MongoClient()
        super().__init__(client=client, database=database)

    def get_collection(self, model: Type[ModelType]) -> "Collection":
        """Get the pymongo collection associated to a Model.

        Args:
            model: model class

        Returns:
            the pymongo collection object
        """
        collection = self.database[model.__collection__]
        return collection

    @staticmethod
    def _get_session(
        session: Union[SyncSessionType, SyncSessionBase],
    ) -> Optional[ClientSession]:
        if isinstance(session, (SyncSession, SyncTransaction)):
            return session.get_driver_session()
        assert not isinstance(session, SyncSessionBase)  # Abstract class
        return session

    def configure_database(
        self,
        models: Sequence[Type[ModelType]],
        *,
        update_existing_indexes: bool = False,
        session: SyncSessionType = None,
    ) -> None:
        """Apply model constraints to the database.

        Args:
            models: list of models to initialize the database with
            update_existing_indexes: conflicting indexes will be dropped before creation
            session: an optional session to use for the operation

        <!---
        #noqa: DAR401 pymongo.errors.OperationFailure
        -->
        """
        driver_session = self._get_session(session)
        for model in models:
            collection = self.get_collection(model)
            for index in model.__indexes__():
                pymongo_index = (
                    index.get_pymongo_index()
                    if isinstance(index, ODMBaseIndex)
                    else index
                )
                try:
                    collection.create_indexes([pymongo_index], session=driver_session)
                except pymongo.errors.OperationFailure as exc:
                    if update_existing_indexes and getattr(exc, "code", None) in (
                        85,  # aka IndexOptionsConflict
                        86,  # aka IndexKeySpecsConflict for MongoDB > 5
                    ):
                        collection.drop_index(
                            pymongo_index.document["name"], session=driver_session
                        )
                        collection.create_indexes(
                            [pymongo_index], session=driver_session
                        )
                    else:
                        raise

    def session(self) -> SyncSession:
        """Get a new session for the engine to allow ordering sequential operations.

        Returns:
            a new session object

        Example usage:

        ```python
        engine = SyncEngine(...)
        with engine.session() as session:
            john = session.find(User, User.name == "John")
            john.name = "Doe"
            session.save(john)
        ```
        """
        return SyncSession(self)

    def transaction(self) -> SyncTransaction:
        """Get a new transaction for the engine to aggregate sequential operations.

        Returns:
            a new transaction object

        Example usage:
        ```python
        engine = SyncEngine(...)
        with engine.transaction() as transaction:
            john = transaction.find(User, User.name == "John")
            john.name = "Doe"
            transaction.save(john)
            transaction.commit()
        ```

        Warning:
            MongoDB transaction are only supported on replicated clusters: either
            directly a replicaSet or a sharded cluster with replication enabled.
        """
        return SyncTransaction(self)

    def find(
        self,
        model: Type[ModelType],
        *queries: Union[
            QueryExpression, Dict, bool
        ],  # bool: allow using binary operators with mypy
        sort: Optional[Any] = None,
        skip: int = 0,
        limit: Optional[int] = None,
        session: SyncSessionType = None,
    ) -> SyncCursor[ModelType]:
        """Search for Model instances matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            sort: sort expression
            skip: number of document to skip
            limit: maximum number of instance fetched
            session: an optional session to use for the operation

        Raises:
            DocumentParsingError: unable to parse one of the resulting documents

        Returns:
            [odmantic.engine.SyncCursor][] of the query

        <!---
        #noqa: DAR401 ValueError
        #noqa: DAR401 TypeError
        #noqa: DAR402 DocumentParsingError
        -->
        """
        pipeline = self._prepare_find_pipeline(
            model,
            *queries,
            sort=sort,
            skip=skip,
            limit=limit,
        )
        collection = self.get_collection(model)
        cursor = collection.aggregate(pipeline, session=self._get_session(session))
        return SyncCursor(model, cursor)

    def find_one(
        self,
        model: Type[ModelType],
        *queries: Union[
            QueryExpression, Dict, bool
        ],  # bool: allow using binary operators w/o plugin
        sort: Optional[Any] = None,
        session: SyncSessionType = None,
    ) -> Optional[ModelType]:
        """Search for a Model instance matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            sort: sort expression
            session: an optional session to use for the operation

        Raises:
            DocumentParsingError: unable to parse the resulting document

        Returns:
            the fetched instance if found otherwise None

        <!---
        #noqa: DAR401 TypeError
        #noqa: DAR402 DocumentParsingError
        -->
        """
        if not lenient_issubclass(model, Model):
            raise TypeError("Can only call find_one with a Model class")
        results = list(self.find(model, *queries, sort=sort, limit=1, session=session))
        if len(results) == 0:
            return None
        return results[0]

    def _save(self, instance: ModelType, session: "ClientSession") -> ModelType:
        """Perform an atomic save operation in the specified session"""
        for ref_field_name in instance.__references__:
            sub_instance = cast(Model, getattr(instance, ref_field_name))
            self._save(sub_instance, session)

        fields_to_update = instance.__fields_modified__ | instance.__mutable_fields__
        if len(fields_to_update) > 0:
            doc = instance.model_dump_doc(include=fields_to_update)
            collection = self.get_collection(type(instance))
            try:
                collection.update_one(
                    instance.model_dump_doc(include={instance.__primary_field__}),
                    {"$set": doc},
                    upsert=True,
                    session=session,
                )
            except pymongo.errors.DuplicateKeyError as e:
                raise DuplicateKeyError(instance, e)
            object.__setattr__(instance, "__fields_modified__", set())
        return instance

    def save(
        self,
        instance: ModelType,
        *,
        session: SyncSessionType = None,
    ) -> ModelType:
        """Persist an instance to the database

        This method behaves as an 'upsert' operation. If a document already exists
        with the same primary key, it will be overwritten.

        All the other models referenced by this instance will be saved as well.

        Args:
            instance: instance to persist
            session: An optional session to use for the operation. If not provided, an
                     internal session will be used to persist the instance and
                     sub-instances.

        Returns:
            the saved instance

        Raises:
            DuplicateKeyError: the instance is duplicated according to a unique index.

        NOTE:
            The save operation actually modify the instance argument in place. However,
            the instance is still returned for convenience.

        <!---
        #noqa: DAR401 TypeError
        #noqa: DAR402 DuplicateKeyError
        -->
        """
        if not isinstance(instance, Model):
            raise TypeError("Can only call find_one with a Model class")

        if session:
            self._save(instance, self._get_session(session))  # type: ignore
        else:
            with self.client.start_session() as local_session:
                self._save(instance, local_session)
        return instance

    def save_all(
        self,
        instances: Sequence[ModelType],
        *,
        session: SyncSessionType = None,
    ) -> List[ModelType]:
        """Persist instances to the database

        This method behaves as multiple 'upsert' operations. If one of the document
        already exists with the same primary key, it will be overwritten.

        All the other models referenced by this instance will be recursively saved as
        well.

        Args:
            instances: instances to persist
            session: An optional session to use for the operation. If not provided an
                     internal session will be used to persist the instances.

        Returns:
            the saved instances

        Raises:
            DuplicateKeyError: an instance is duplicated according to a unique index.

        NOTE:
            The save_all operation actually modify the arguments in place. However, the
            instances are still returned for convenience.
        <!---
        #noqa: DAR402 DuplicateKeyError
        -->
        """
        if session:
            added_instances = [
                self._save(instance, self._get_session(session))  # type: ignore
                for instance in instances
            ]
        else:
            with self.client.start_session() as local_session:
                added_instances = [
                    self._save(instance, local_session) for instance in instances
                ]
        return added_instances

    def delete(
        self,
        instance: ModelType,
        session: SyncSessionType = None,
    ) -> None:
        """Delete an instance from the database

        Args:
            instance: the instance to delete
            session: an optional session to use for the operation

        Raises:
            DocumentNotFoundError: the instance has not been persisted to the database

        """
        # TODO handle cascade deletion
        collection = self.database[instance.__collection__]
        pk_name = instance.__primary_field__
        result = collection.delete_many(
            {"_id": getattr(instance, pk_name)}, session=self._get_session(session)
        )
        count = result.deleted_count
        if count == 0:
            raise DocumentNotFoundError(instance)

    def remove(
        self,
        model: Type[ModelType],
        *queries: Union[QueryExpression, Dict, bool],
        just_one: bool = False,
        session: SyncSessionType = None,
    ) -> int:
        """Delete Model instances matching the query filter provided

        Args:
            model: model to perform the operation on
            *queries: query filter to apply
            just_one: limit the deletion to just one document
            session: an optional session to use for the operation

        Returns:
            the number of instances deleted from the database.
        """
        query = SyncEngine._build_query(*queries)
        collection = self.get_collection(model)

        if just_one:
            result = collection.delete_one(query, session=self._get_session(session))
        else:
            result = collection.delete_many(query, session=self._get_session(session))

        return result.deleted_count

    def count(
        self,
        model: Type[ModelType],
        *queries: Union[QueryExpression, Dict, bool],
        session: SyncSessionType = None,
    ) -> int:
        """Get the count of documents matching a query

        Args:
            model: model to perform the operation on
            *queries: query filters to apply
            session: an optional session to use for the operation

        Returns:
            number of document matching the query

        <!---
        #noqa: DAR401 TypeError
        -->
        """
        if not lenient_issubclass(model, Model):
            raise TypeError("Can only call count with a Model class")
        query = BaseEngine._build_query(*queries)
        collection = self.database[model.__collection__]
        count = collection.count_documents(query, session=self._get_session(session))
        return int(count)



from bson.binary import Binary
from bson.decimal128 import Decimal128
from bson.int64 import Int64
from bson.objectid import ObjectId
from bson.regex import Regex

from odmantic.model import Model


class FullBsonModel(Model):
    """Model used to test BSON types"""

    objectId_: ObjectId
    long_: Int64
    decimal_: Decimal128
    binary_: Binary
    regex_: Regex



import sys
from unittest.mock import MagicMock, Mock

import pytest

from odmantic.session import AIOSession, AIOTransaction, SyncSession, SyncTransaction

from ..zoo.person import PersonModel

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock


pytestmark = [
    pytest.mark.asyncio,
    pytest.mark.skipif(
        sys.version_info < (3, 8), reason="async mock testing requires python3.8+"
    ),
]


@pytest.fixture(scope="function")
def mocked_driver_session():
    return Mock()


@pytest.fixture(scope="function")
def mocked_aio_engine(mocked_driver_session):
    engine = AsyncMock()
    engine._get_session = Mock(return_value=mocked_driver_session)
    return engine


@pytest.fixture(scope="function", params=["session", "transaction"])
async def mocked_aio_session(request, mocked_aio_engine):
    if request.param == "session":
        return AIOSession(mocked_aio_engine)
    else:
        session = AIOSession(mocked_aio_engine)
        await session.start()
        return AIOTransaction(session)


@pytest.fixture(scope="function")
def mocked_sync_engine(mocked_driver_session):
    engine = MagicMock()
    engine._get_session = Mock(return_value=mocked_driver_session)
    return engine


@pytest.fixture(scope="function", params=["session", "transaction"])
def mocked_sync_session(request, mocked_sync_engine):
    if request.param == "session":
        return SyncSession(mocked_sync_engine)
    else:
        session = SyncSession(mocked_sync_engine)
        session.start()
        return SyncTransaction(session)


async def test_session_find(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.find(PersonModel)
    mocked_aio_engine.find.assert_awaited_once()
    assert mocked_aio_engine.find.call_args.kwargs["session"] == mocked_driver_session


async def test_session_find_one(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.find_one(PersonModel)
    mocked_aio_engine.find_one.assert_awaited_once()
    assert (
        mocked_aio_engine.find_one.call_args.kwargs["session"] == mocked_driver_session
    )


async def test_session_count(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.count(PersonModel)
    mocked_aio_engine.count.assert_awaited_once()
    assert mocked_aio_engine.count.call_args.kwargs["session"] == mocked_driver_session


async def test_session_save(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.save(PersonModel(first_name="John", last_name="Doe"))
    mocked_aio_engine.save.assert_awaited_once()
    assert mocked_aio_engine.save.call_args.kwargs["session"] == mocked_driver_session


async def test_session_save_all(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.save_all([PersonModel(first_name="John", last_name="Doe")])
    mocked_aio_engine.save_all.assert_awaited_once()
    assert (
        mocked_aio_engine.save_all.call_args.kwargs["session"] == mocked_driver_session
    )


async def test_session_delete(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.delete(PersonModel(first_name="John", last_name="Doe"))
    mocked_aio_engine.delete.assert_awaited_once()
    assert mocked_aio_engine.delete.call_args.kwargs["session"] == mocked_driver_session


async def test_session_remove(
    mocked_aio_session, mocked_aio_engine, mocked_driver_session
):
    await mocked_aio_session.remove(PersonModel, PersonModel.first_name == "John")
    mocked_aio_engine.remove.assert_awaited_once()
    assert mocked_aio_engine.remove.call_args.kwargs["session"] == mocked_driver_session


def test_sync_session_find(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.find(PersonModel)
    mocked_sync_engine.find.assert_called_once()
    assert mocked_sync_engine.find.call_args.kwargs["session"] == mocked_driver_session


def test_sync_session_find_one(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.find_one(PersonModel)
    mocked_sync_engine.find_one.assert_called_once()
    assert (
        mocked_sync_engine.find_one.call_args.kwargs["session"] == mocked_driver_session
    )


def test_sync_session_count(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.count(PersonModel)
    mocked_sync_engine.count.assert_called_once()
    assert mocked_sync_engine.count.call_args.kwargs["session"] == mocked_driver_session


def test_sync_session_save(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.save(PersonModel(first_name="John", last_name="Doe"))
    mocked_sync_engine.save.assert_called_once()
    assert mocked_sync_engine.save.call_args.kwargs["session"] == mocked_driver_session


def test_sync_session_save_all(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.save_all([PersonModel(first_name="John", last_name="Doe")])
    mocked_sync_engine.save_all.assert_called_once()
    assert (
        mocked_sync_engine.save_all.call_args.kwargs["session"] == mocked_driver_session
    )


def test_sync_session_delete(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.delete(PersonModel(first_name="John", last_name="Doe"))
    mocked_sync_engine.delete.assert_called_once()
    assert (
        mocked_sync_engine.delete.call_args.kwargs["session"] == mocked_driver_session
    )


def test_sync_session_remove(
    mocked_sync_session, mocked_sync_engine, mocked_driver_session
):
    mocked_sync_session.remove(PersonModel, PersonModel.first_name == "John")
    mocked_sync_engine.remove.assert_called_once()
    assert (
        mocked_sync_engine.remove.call_args.kwargs["session"] == mocked_driver_session
    )



import pymongo

from odmantic import Model


class Post(Model):
    title: str
    content: str

    model_config = {
        "indexes": lambda: [
            pymongo.IndexModel(
                [(+Post.title, pymongo.TEXT), (+Post.content, pymongo.TEXT)]
            )
        ]
    }



import pymongo
import pytest

from odmantic.engine import AIOEngine, SyncEngine
from odmantic.exceptions import DuplicateKeyError
from odmantic.field import Field
from odmantic.index import Index
from odmantic.model import Model
from odmantic.query import asc, desc

pytestmark = pytest.mark.asyncio


async def test_single_field_index_creation(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(index=True)

    await aio_engine.configure_database([M])

    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )


def test_sync_single_field_index_creation(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(index=True)

    sync_engine.configure_database([M])

    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )


async def test_single_field_index_creation_unique(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(unique=True)

    await aio_engine.configure_database([M])

    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )


def test_sync_single_field_index_creation_unique(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(unique=True)

    sync_engine.configure_database([M])

    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )


async def test_compound_index_with_name(aio_engine: AIOEngine):
    class M(Model):
        f: int
        g: int

        model_config = {"indexes": lambda: [Index(asc(M.f), desc(M.g), name="test")]}

    await aio_engine.configure_database([M])

    info = await aio_engine.get_collection(M).index_information()
    assert "test" in info
    assert info["test"]["key"] == [("f", 1), ("g", -1)]


def test_sync_compound_index_with_name(sync_engine: SyncEngine):
    class M(Model):
        f: int
        g: int

        model_config = {"indexes": lambda: [Index(asc(M.f), desc(M.g), name="test")]}

    sync_engine.configure_database([M])

    info = sync_engine.get_collection(M).index_information()
    assert "test" in info
    assert info["test"]["key"] == [("f", 1), ("g", -1)]


async def test_multiple_indexes(aio_engine: AIOEngine):
    class M(Model):
        f: int
        g: int = Field(unique=True)

        model_config = {
            "indexes": lambda: [Index(asc(M.f)), Index(asc(M.f), desc(M.g))]
        }

    await aio_engine.configure_database([M])

    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )
    assert (
        next(
            filter(lambda v: v["key"] == [("g", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1), ("g", -1)], info.values()),
            None,
        )
        is not None
    )


def test_sync_multiple_indexes(sync_engine: SyncEngine):
    class M(Model):
        f: int
        g: int = Field(unique=True)

        model_config = {
            "indexes": lambda: [Index(asc(M.f)), Index(asc(M.f), desc(M.g))]
        }

    sync_engine.configure_database([M])

    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )
    assert (
        next(
            filter(lambda v: v["key"] == [("g", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1), ("g", -1)], info.values()),
            None,
        )
        is not None
    )


async def test_unique_index_duplicate_save(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(unique=True)

    await aio_engine.configure_database([M])

    await aio_engine.save(M(f=1))
    duplicated_instance = M(f=1)
    with pytest.raises(DuplicateKeyError) as e:
        await aio_engine.save(duplicated_instance)
    assert e.value.instance == duplicated_instance


def test_sync_unique_index_duplicate_save(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(unique=True)

    sync_engine.configure_database([M])

    sync_engine.save(M(f=1))
    duplicated_instance = M(f=1)
    with pytest.raises(DuplicateKeyError) as e:
        sync_engine.save(duplicated_instance)
    assert e.value.instance == duplicated_instance


async def test_double_index_creation(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(index=True)

    await aio_engine.configure_database([M])
    await aio_engine.configure_database([M])

    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )


def test_sync_double_index_creation(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(index=True)

    sync_engine.configure_database([M])
    sync_engine.configure_database([M])

    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)], info.values()),
            None,
        )
        is not None
    )


async def test_index_update_failure(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(unique=True)

        model_config = {
            "collection": "test",
        }

    await aio_engine.configure_database([M])

    class M2(Model):
        f: int = Field(index=True)

        model_config = {
            "collection": "test",
        }

    with pytest.raises(pymongo.errors.OperationFailure):
        await aio_engine.configure_database([M2])


def test_sync_index_update_failure(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(unique=True)

        model_config = {
            "collection": "test",
        }

    sync_engine.configure_database([M])

    class M2(Model):
        f: int = Field(index=True)

        model_config = {
            "collection": "test",
        }

    with pytest.raises(pymongo.errors.OperationFailure):
        sync_engine.configure_database([M2])


async def test_index_replacement(aio_engine: AIOEngine):
    class M(Model):
        f: int = Field(unique=True)

        model_config = {
            "collection": "test",
        }

    await aio_engine.configure_database([M])
    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )

    class M2(Model):
        f: int = Field(index=True)

        model_config = {
            "collection": "test",
        }

    await aio_engine.configure_database([M2], update_existing_indexes=True)
    info = await aio_engine.get_collection(M).index_information()
    assert (
        next(
            filter(
                lambda v: v["key"] == [("f", 1)] and "unique" not in v, info.values()
            ),
            None,
        )
        is not None
    )


def test_sync_index_replacement(sync_engine: SyncEngine):
    class M(Model):
        f: int = Field(unique=True)

        model_config = {
            "collection": "test",
        }

    sync_engine.configure_database([M])
    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(lambda v: v["key"] == [("f", 1)] and v["unique"], info.values()),
            None,
        )
        is not None
    )

    class M2(Model):
        f: int = Field(index=True)

        model_config = {
            "collection": "test",
        }

    sync_engine.configure_database([M2], update_existing_indexes=True)
    info = sync_engine.get_collection(M).index_information()
    assert (
        next(
            filter(
                lambda v: v["key"] == [("f", 1)] and "unique" not in v, info.values()
            ),
            None,
        )
        is not None
    )


async def test_custom_text_index(aio_engine: AIOEngine):
    class Post(Model):
        title: str
        content: str

        model_config = {
            "indexes": lambda: [
                pymongo.IndexModel([("title", pymongo.TEXT), ("content", pymongo.TEXT)])
            ]
        }

    await aio_engine.configure_database([Post])
    await aio_engine.save(Post(title="My post on python", content="It's awesome!"))
    assert await aio_engine.find_one(Post, {"$text": {"$search": "python"}}) is not None


async def test_sync_custom_text_index(sync_engine: SyncEngine):
    class Post(Model):
        title: str
        content: str

        model_config = {
            "indexes": lambda: [
                pymongo.IndexModel([("title", pymongo.TEXT), ("content", pymongo.TEXT)])
            ]
        }

    sync_engine.configure_database([Post])
    sync_engine.save(Post(title="My post on python", content="It's awesome!"))
    assert sync_engine.find_one(Post, {"$text": {"$search": "python"}}) is not None



from odmantic import Model, query


class Tree(Model):
    name: str
    average_size: float


Tree.name == "Spruce"
#> QueryExpression({'name': {'$eq': 'Spruce'}})
Tree.name.eq("Spruce")
#> QueryExpression({'name': {'$eq': 'Spruce'}})
query.eq(Tree.name, "Spruce")
#> QueryExpression({'name': {'$eq': 'Spruce'}})



from typing import Type

from odmantic.typing import get_args, get_origin


def are_generics_equal(g1: Type, g2: Type) -> bool:
    """Check if two generic types are equal."""
    if g1 == g2:
        return True
    origin_g1 = get_origin(g1)
    origin_g2 = get_origin(g2)
    if origin_g1 is None or origin_g2 is None or origin_g1 != origin_g2:
        return False
    args_g1 = get_args(g1)
    args_g2 = get_args(g2)
    if len(args_g1) != len(args_g2):
        return False
    return all(are_generics_equal(a1, a2) for a1, a2 in zip(args_g1, args_g2))



from odmantic import Field, Model


class Player(Model):
    name: str
    level: int = Field(default=1, ge=1)


p = Player(name="Dash")
print(repr(p))
#> Player(id=ObjectId('5f7cdbfbb54a94e9e8717c77'), name='Dash', level=1)



from datetime import datetime
from typing import Optional

import pytest

import odmantic
from odmantic.field import Field
from odmantic.model import EmbeddedModel, Model
from odmantic.reference import Reference


def test_field_defined_as_primary_key_and_custom_name():
    with pytest.raises(
        ValueError, match="cannot specify a primary field with a custom key_name"
    ):
        Field(primary_field=True, key_name="not _id")


def test_field_defined_as_primary_key_default_name():
    f = Field(primary_field=True)
    assert f.key_name == "_id"


def test_field_define_key_as__id_without_setting_as_primary():
    with pytest.raises(
        ValueError,
        match="cannot specify key_name='_id' without defining the field as primary",
    ):
        Field(key_name="_id")


def test_pos_key_name():
    class M(Model):
        field: int = Field(key_name="alternate_name")

    assert +M.field == "alternate_name"
    assert ++M.field == "$alternate_name"


def test_unknown_attr_embedded_model():
    class E(EmbeddedModel): ...

    class M(Model):
        field: E

    with pytest.raises(AttributeError):
        M.field.unknown_attr  # type: ignore


@pytest.mark.parametrize("operator_name", ("lt", "lte", "gt", "gte", "match"))
def test_reference_field_operator_not_allowed(operator_name: str):
    class E(Model): ...

    class M(Model):
        field: E = Reference()

    with pytest.raises(
        AttributeError,
        match=f"operator {operator_name} not allowed for ODMReference fields",
    ):
        getattr(M.field, operator_name)


def test_field_required_in_doc_without_default():
    class M(Model):
        field: str

    assert M.__odm_fields__["field"].is_required_in_doc()


def test_field_required_in_doc_with_default():
    class M(Model):
        field: str = Field("hi")

    assert not M.__odm_fields__["field"].is_required_in_doc()


def test_field_required_in_doc_default_factory_disabled():
    class M(Model):
        field: str = Field(default_factory=lambda: "hi")  # pragma: no cover

    assert M.__odm_fields__["field"].is_required_in_doc()


def test_field_required_in_doc_default_factory_enabled():
    class M(Model):
        field: str = Field(default_factory=lambda: "hi")  # pragma: no cover

        model_config = {
            "parse_doc_with_default_factories": True,
        }

    assert not M.__odm_fields__["field"].is_required_in_doc()


def test_multiple_optional_fields():
    class M(Model):
        field: str = Field(default_factory=lambda: "hi")  # pragma: no cover
        optionalBoolField: Optional[bool] = None
        optionalDatetimeField: Optional[datetime] = None

    assert (
        M.__odm_fields__["optionalBoolField"].pydantic_field.annotation
        == Optional[bool]
    )
    assert (
        M.__odm_fields__["optionalDatetimeField"].pydantic_field.annotation
        == Optional[odmantic.bson._datetime]
    )

    instance = M(field="Hi")
    # This should work and never throw a value error
    instance.optionalBoolField = True



import pytest

from odmantic import EmbeddedModel, Model


class M(Model): ...


def test_deprecated_copy():
    with pytest.deprecated_call():
        M().copy()


def test_deprecated_update():
    with pytest.deprecated_call():
        M().update({})


def test_deprecated_update_basemodel():
    # EmbeddedModel is a subclass of BaseModel not redefine update
    class E(EmbeddedModel): ...

    with pytest.deprecated_call():
        E().update({})


def test_deprecated_doc():
    with pytest.deprecated_call():
        M().doc()


def test_deprecated_parse_doc():
    with pytest.deprecated_call():
        M().parse_doc(
            {
                "_id": "5f8352a87a733b8b18b0cb27",
            }
        )



from typing import Dict, List, Tuple

import pytest

from odmantic.engine import AIOEngine, SyncEngine
from odmantic.field import Field
from odmantic.model import EmbeddedModel, Model

from ..zoo.book_embedded import Book, Publisher
from ..zoo.patron_embedded import Address, Patron

pytestmark = pytest.mark.asyncio


async def test_add_fetch_single(aio_engine: AIOEngine):
    publisher = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    book = Book(title="MongoDB: The Definitive Guide", pages=216, publisher=publisher)
    instance = await aio_engine.save(book)
    assert instance.id is not None
    assert isinstance(instance.publisher, Publisher)
    assert instance.publisher == publisher

    fetched_instance = await aio_engine.find_one(Book, Book.id == instance.id)
    assert fetched_instance is not None
    assert fetched_instance.publisher == publisher


def test_sync_add_fetch_single(sync_engine: SyncEngine):
    publisher = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    book = Book(title="MongoDB: The Definitive Guide", pages=216, publisher=publisher)
    instance = sync_engine.save(book)
    assert instance.id is not None
    assert isinstance(instance.publisher, Publisher)
    assert instance.publisher == publisher

    fetched_instance = sync_engine.find_one(Book, Book.id == instance.id)
    assert fetched_instance is not None
    assert fetched_instance.publisher == publisher


async def test_add_multiple(aio_engine: AIOEngine):
    addresses = [
        Address(street="81 Lafayette St.", city="Brownsburg", state="IN", zip="46112"),
        Address(
            street="862 West Euclid St.", city="Indian Trail", state="NC", zip="28079"
        ),
    ]
    patron = Patron(name="The Princess Royal", addresses=addresses)
    instance = await aio_engine.save(patron)
    assert instance.id is not None
    assert isinstance(instance.addresses, list)
    assert instance.addresses == addresses

    fetched_instance = await aio_engine.find_one(Patron)
    assert fetched_instance is not None
    assert fetched_instance.addresses == addresses


def test_sync_add_multiple(sync_engine: SyncEngine):
    addresses = [
        Address(street="81 Lafayette St.", city="Brownsburg", state="IN", zip="46112"),
        Address(
            street="862 West Euclid St.", city="Indian Trail", state="NC", zip="28079"
        ),
    ]
    patron = Patron(name="The Princess Royal", addresses=addresses)
    instance = sync_engine.save(patron)
    assert instance.id is not None
    assert isinstance(instance.addresses, list)
    assert instance.addresses == addresses

    fetched_instance = sync_engine.find_one(Patron)
    assert fetched_instance is not None
    assert fetched_instance.addresses == addresses


@pytest.fixture
async def books_with_embedded_publisher(aio_engine: AIOEngine):
    publisher_1 = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    book_1 = Book(
        title="MongoDB: The Definitive Guide", pages=216, publisher=publisher_1
    )
    publisher_2 = Publisher(name="O'Reilly Media", founded=2020, location="EU")
    book_2 = Book(title="MySQL: The Definitive Guide", pages=516, publisher=publisher_2)
    return await aio_engine.save_all([book_1, book_2])


async def test_query_filter_on_embedded_doc(
    aio_engine: AIOEngine, books_with_embedded_publisher: Tuple[Book, Book]
):
    _, book_2 = books_with_embedded_publisher
    fetched_instances = await aio_engine.find(Book, Book.publisher == book_2.publisher)
    assert len(fetched_instances) == 1
    assert fetched_instances[0] == book_2


def test_sync_query_filter_on_embedded_doc(
    sync_engine: SyncEngine, books_with_embedded_publisher: Tuple[Book, Book]
):
    _, book_2 = books_with_embedded_publisher
    fetched_instances = list(sync_engine.find(Book, Book.publisher == book_2.publisher))
    assert len(fetched_instances) == 1
    assert fetched_instances[0] == book_2


async def test_query_filter_on_embedded_field(
    aio_engine: AIOEngine, books_with_embedded_publisher: Tuple[Book, Book]
):
    _, book_2 = books_with_embedded_publisher
    fetched_instances = await aio_engine.find(Book, Book.publisher.location == "EU")
    assert len(fetched_instances) == 1
    assert fetched_instances[0] == book_2


def test_sync_query_filter_on_embedded_field(
    sync_engine: SyncEngine, books_with_embedded_publisher: Tuple[Book, Book]
):
    _, book_2 = books_with_embedded_publisher
    fetched_instances = list(sync_engine.find(Book, Book.publisher.location == "EU"))
    assert len(fetched_instances) == 1
    assert fetched_instances[0] == book_2


async def test_query_filter_on_embedded_nested(aio_engine: AIOEngine):
    class ThirdModel(EmbeddedModel):
        field: int

    class SecondaryModel(EmbeddedModel):
        nested_1: ThirdModel

    class TopModel(Model):
        nested_0: SecondaryModel

    instance_0 = TopModel(nested_0=SecondaryModel(nested_1=ThirdModel(field=12)))
    instance_1 = TopModel(nested_0=SecondaryModel(nested_1=ThirdModel(field=0)))
    await aio_engine.save_all([instance_0, instance_1])

    fetched_instances = await aio_engine.find(
        TopModel, TopModel.nested_0.nested_1.field == 12
    )

    assert len(fetched_instances) == 1
    assert fetched_instances[0] == instance_0


def test_sync_query_filter_on_embedded_nested(sync_engine: SyncEngine):
    class ThirdModel(EmbeddedModel):
        field: int

    class SecondaryModel(EmbeddedModel):
        nested_1: ThirdModel

    class TopModel(Model):
        nested_0: SecondaryModel

    instance_0 = TopModel(nested_0=SecondaryModel(nested_1=ThirdModel(field=12)))
    instance_1 = TopModel(nested_0=SecondaryModel(nested_1=ThirdModel(field=0)))
    sync_engine.save_all([instance_0, instance_1])

    fetched_instances = list(
        sync_engine.find(TopModel, TopModel.nested_0.nested_1.field == 12)
    )

    assert len(fetched_instances) == 1
    assert fetched_instances[0] == instance_0


async def test_fields_modified_embedded_model_modification(aio_engine: AIOEngine):
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

    e = E(f=0)
    m = M(e=e)
    await aio_engine.save(m)
    e.f = 1
    await aio_engine.save(m)
    fetched = await aio_engine.find_one(M)
    assert fetched is not None
    assert fetched.e.f == 1


def test_sync_fields_modified_embedded_model_modification(sync_engine: SyncEngine):
    class E(EmbeddedModel):
        f: int

    class M(Model):
        e: E

    e = E(f=0)
    m = M(e=e)
    sync_engine.save(m)
    e.f = 1
    sync_engine.save(m)
    fetched = sync_engine.find_one(M)
    assert fetched is not None
    assert fetched.e.f == 1


async def test_embedded_model_as_primary_field_named_id(aio_engine: AIOEngine):
    class Id(EmbeddedModel):
        user: int
        chat: int

    class User(Model):
        id: Id = Field(primary_field=True)

    user = User(id=Id(user=1, chat=1001))
    await aio_engine.save(user)
    assert (
        await aio_engine.find_one(User, User.id == Id(user=1, chat=1001))
    ) is not None


async def test_sync_embedded_model_as_primary_field_named_id(sync_engine: SyncEngine):
    class Id(EmbeddedModel):
        user: int
        chat: int

    class User(Model):
        id: Id = Field(primary_field=True)

    user = User(id=Id(user=1, chat=1001))
    sync_engine.save(user)
    assert sync_engine.find_one(User, User.id == Id(user=1, chat=1001)) is not None


async def test_embedded_model_custom_key_name_save_and_fetch(aio_engine: AIOEngine):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: In = Field(key_name="in")

    instance = Out(inner=In(a=3))
    await aio_engine.save(instance)
    fetched = await aio_engine.find_one(Out)
    assert instance == fetched


def test_sync_embedded_model_custom_key_name_save_and_fetch(sync_engine: SyncEngine):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: In = Field(key_name="in")

    instance = Out(inner=In(a=3))
    sync_engine.save(instance)
    fetched = sync_engine.find_one(Out)
    assert instance == fetched


async def test_embedded_model_list_custom_key_name_save_and_fetch(
    aio_engine: AIOEngine,
):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: List[In] = Field(key_name="in")

    instance = Out(inner=[In(a=3)])
    await aio_engine.save(instance)
    fetched = await aio_engine.find_one(Out)
    assert instance == fetched


def test_sync_embedded_model_list_custom_key_name_save_and_fetch(
    sync_engine: SyncEngine,
):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: List[In] = Field(key_name="in")

    instance = Out(inner=[In(a=3)])
    sync_engine.save(instance)
    fetched = sync_engine.find_one(Out)
    assert instance == fetched


async def test_embedded_model_dict_custom_key_name_save_and_fetch(
    aio_engine: AIOEngine,
):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: Dict[str, In] = Field(key_name="in")

    instance = Out(inner={"key": In(a=3)})
    await aio_engine.save(instance)
    fetched = await aio_engine.find_one(Out)
    assert instance == fetched


def test_sync_embedded_model_dict_custom_key_name_save_and_fetch(
    sync_engine: SyncEngine,
):
    class In(EmbeddedModel):
        a: int = Field(key_name="in-a")

    class Out(Model):
        inner: Dict[str, In] = Field(key_name="in")

    instance = Out(inner={"key": In(a=3)})
    sync_engine.save(instance)
    fetched = sync_engine.find_one(Out)
    assert instance == fetched



from bson import ObjectId

from odmantic import SyncEngine, Model


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

shroud = engine.find_one(Player, Player.name == "Shroud")
print(shroud.id)
#> 5f86074f6dfecacc68428a62
new_id = ObjectId("ffffffffffffffffffffffff")
# Copy the player instance with a new primary key
new_shroud = shroud.copy(update={"id": new_id})
# Delete the initial player instance
engine.delete(shroud)
# Finally persist again the new instance
engine.save(new_shroud)



import sys
from typing import TYPE_CHECKING, AbstractSet, Any  # noqa: F401
from typing import Callable as TypingCallable
from typing import Dict, Iterable, Mapping, Tuple, Type, TypeVar, Union  # noqa: F401

from pydantic.v1.typing import is_classvar, resolve_annotations  # noqa: F401
from pydantic.v1.utils import lenient_issubclass

if sys.version_info < (3, 11):
    from typing_extensions import dataclass_transform
else:
    from typing import dataclass_transform  # noqa: F401

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if sys.version_info < (3, 9):
    from typing import _GenericAlias as GenericAlias  # noqa: F401

    # Even if get_args and get_origin are available in typing, it's important to
    # import them from typing_extensions to have proper origins with Annotated fields
    from typing_extensions import Annotated, get_args, get_origin
else:
    from typing import GenericAlias  # type: ignore  # noqa: F401
    from typing import Annotated, get_args, get_origin  # noqa: F401


if TYPE_CHECKING:
    NoArgAnyCallable: TypeAlias = TypingCallable[[], Any]
    ReprArgs: TypeAlias = "Iterable[tuple[str | None, Any]]"
    AbstractSetIntStr: TypeAlias = "AbstractSet[int] | AbstractSet[str]"
    MappingIntStrAny: TypeAlias = "Mapping[int, Any] | Mapping[str, Any]"
    DictStrAny: TypeAlias = Dict[str, Any]
    IncEx: TypeAlias = "set[int] | set[str] | dict[int, Any] | dict[str, Any] | None"


def is_type_argument_subclass(
    type_: Type, class_or_tuple: Union[Type[Any], Tuple[Type[Any], ...]]
) -> bool:
    args = get_args(type_)
    return any(lenient_issubclass(arg, class_or_tuple) for arg in args)


T = TypeVar("T")


def get_first_type_argument_subclassing(
    type_: Type, cls: Type[T]
) -> Union[Type[T], None]:
    args: Tuple[Type, ...] = get_args(type_)
    for arg in args:
        if lenient_issubclass(arg, cls):
            return arg
    return None



from enum import Enum

from odmantic import Model, query


class TreeKind(str, Enum):
    BIG = "big"
    SMALL = "small"


class Tree(Model):
    name: str
    average_size: float
    kind: TreeKind


Tree.kind == TreeKind.SMALL
#> QueryExpression({'kind': {'$eq': 'small'}})
Tree.kind.eq(TreeKind.SMALL)
#> QueryExpression({'kind': {'$eq': 'small'}})
query.eq(Tree.kind, TreeKind.SMALL)
#> QueryExpression({'kind': {'$eq': 'small'}})



from odmantic import AIOEngine, Model


class User(Model):
    name: str


engine = AIOEngine()
motor_collection = engine.get_collection(User)
print(motor_collection)
#> AsyncIOMotorCollection(
#>     Collection(
#>         Database(
#>             MongoClient(
#>                 host=["localhost:27017"],
#>                 document_class=dict,
#>                 tz_aware=False,
#>                 connect=False,
#>                 driver=DriverInfo(name="Motor", version="2.2.0", platform="asyncio"),
#>             ),
#>             "test",
#>         ),
#>         "user",
#>     )
#> )



import pytest
from inline_snapshot import snapshot

from odmantic.bson import ObjectId
from odmantic.engine import AIOEngine, SyncEngine
from odmantic.exceptions import DocumentParsingError
from odmantic.model import Model
from odmantic.reference import Reference
from tests.integration.conftest import only_on_replica
from tests.integration.utils import redact_objectid
from tests.zoo.deeply_nested import NestedLevel1, NestedLevel2, NestedLevel3

from ..zoo.book_reference import Book, Publisher

pytestmark = pytest.mark.asyncio


async def test_add_with_references(aio_engine: AIOEngine):
    publisher = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    book = Book(title="MongoDB: The Definitive Guide", pages=216, publisher=publisher)
    instance = await aio_engine.save(book)
    fetched_subinstance = await aio_engine.find_one(
        Publisher, Publisher.id == instance.publisher.id
    )
    assert fetched_subinstance == publisher


def test_sync_add_with_references(sync_engine: SyncEngine):
    publisher = Publisher(name="O'Reilly Media", founded=1980, location="CA")
    book = Book(title="MongoDB: The Definitive Guide", pages=216, publisher=publisher)
    instance = sync_engine.save(book)
    fetched_subinstance = sync_engine.find_one(
        Publisher, Publisher.id == instance.publisher.id
    )
    assert fetched_subinstance == publisher


# TODO Handle the case where the referenced object already exists

# TODO test add with duplicated reference id


async def test_save_deeply_nested(aio_engine: AIOEngine):
    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3()))
    await aio_engine.save(instance)
    assert await aio_engine.count(NestedLevel3) == 1
    assert await aio_engine.count(NestedLevel2) == 1
    assert await aio_engine.count(NestedLevel1) == 1


def test_sync_save_deeply_nested(sync_engine: SyncEngine):
    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3()))
    sync_engine.save(instance)
    assert sync_engine.count(NestedLevel3) == 1
    assert sync_engine.count(NestedLevel2) == 1
    assert sync_engine.count(NestedLevel1) == 1


async def test_update_deeply_nested(aio_engine: AIOEngine):
    inst3 = NestedLevel3(
        field=0
    )  # Isolate inst3 to make sure it's not internaly copied
    instance = NestedLevel1(next_=NestedLevel2(next_=inst3))
    await aio_engine.save(instance)
    assert await aio_engine.count(NestedLevel3, NestedLevel3.field == 42) == 0
    inst3.field = 42
    await aio_engine.save(instance)
    assert await aio_engine.count(NestedLevel3, NestedLevel3.field == 42) == 1


def test_sync_update_deeply_nested(sync_engine: SyncEngine):
    inst3 = NestedLevel3(
        field=0
    )  # Isolate inst3 to make sure it's not internaly copied
    instance = NestedLevel1(next_=NestedLevel2(next_=inst3))
    sync_engine.save(instance)
    assert sync_engine.count(NestedLevel3, NestedLevel3.field == 42) == 0
    inst3.field = 42
    sync_engine.save(instance)
    assert sync_engine.count(NestedLevel3, NestedLevel3.field == 42) == 1


async def test_save_deeply_nested_and_fetch(aio_engine: AIOEngine):
    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3(field=0)))
    await aio_engine.save(instance)

    fetched = await aio_engine.find_one(NestedLevel1)
    assert fetched == instance


def test_sync_save_deeply_nested_and_fetch(sync_engine: SyncEngine):
    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3(field=0)))
    sync_engine.save(instance)

    fetched = sync_engine.find_one(NestedLevel1)
    assert fetched == instance


@only_on_replica
async def test_save_deeply_nested_and_fetch_with_transaction(aio_engine: AIOEngine):
    # Before MongoDB 4.4 it's necessary to create the collections before trying to use
    # them inside a transaction
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel1).name
    )
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel2).name
    )
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel3).name
    )

    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3(field=0)))
    async with await aio_engine.client.start_session() as session:
        async with session.start_transaction():
            await aio_engine.save(instance, session=session)

    fetched = await aio_engine.find_one(NestedLevel1)
    assert fetched == instance


@only_on_replica
def test_sync_save_deeply_nested_and_fetch_with_transaction(sync_engine: SyncEngine):
    # Before MongoDB 4.4 it's necessary to create the collections before trying to use
    # them inside a transaction
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel1).name
    )
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel2).name
    )
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel3).name
    )

    instance = NestedLevel1(next_=NestedLevel2(next_=NestedLevel3(field=0)))
    with sync_engine.client.start_session() as session:
        with session.start_transaction():
            sync_engine.save(instance, session=session)

    fetched = sync_engine.find_one(NestedLevel1)
    assert fetched == instance


async def test_multiple_save_deeply_nested_and_fetch(aio_engine: AIOEngine):
    instances = [
        NestedLevel1(field=1, next_=NestedLevel2(field=2, next_=NestedLevel3(field=3))),
        NestedLevel1(field=4, next_=NestedLevel2(field=5, next_=NestedLevel3(field=6))),
    ]
    await aio_engine.save_all(instances)

    fetched = await aio_engine.find(NestedLevel1)
    assert len(fetched) == 2
    assert fetched[0] in instances
    assert fetched[1] in instances


@only_on_replica
async def test_multiple_save_deeply_nested_and_fetch_with_transaction(
    aio_engine: AIOEngine,
):
    # Before MongoDB 4.4 it's necessary to create the collections before trying to use
    # them inside a transaction
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel1).name
    )
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel2).name
    )
    await aio_engine.database.create_collection(
        aio_engine.get_collection(NestedLevel3).name
    )
    instances = [
        NestedLevel1(field=1, next_=NestedLevel2(field=2, next_=NestedLevel3(field=3))),
        NestedLevel1(field=4, next_=NestedLevel2(field=5, next_=NestedLevel3(field=6))),
    ]
    async with await aio_engine.client.start_session() as session:
        async with session.start_transaction():
            await aio_engine.save_all(instances, session=session)

    fetched = await aio_engine.find(NestedLevel1)
    assert len(fetched) == 2
    assert fetched[0] in instances
    assert fetched[1] in instances


@only_on_replica
def test_sync_multiple_save_deeply_nested_and_fetch_with_transaction(
    sync_engine: SyncEngine,
):
    # Before MongoDB 4.4 it's necessary to create the collections before trying to use
    # them inside a transaction
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel1).name
    )
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel2).name
    )
    sync_engine.database.create_collection(
        sync_engine.get_collection(NestedLevel3).name
    )
    instances = [
        NestedLevel1(field=1, next_=NestedLevel2(field=2, next_=NestedLevel3(field=3))),
        NestedLevel1(field=4, next_=NestedLevel2(field=5, next_=NestedLevel3(field=6))),
    ]
    with sync_engine.client.start_session() as session:
        with session.start_transaction():
            sync_engine.save_all(instances, session=session)

    fetched = list(sync_engine.find(NestedLevel1))
    assert len(fetched) == 2
    assert fetched[0] in instances
    assert fetched[1] in instances


def test_sync_multiple_save_deeply_nested_and_fetch(sync_engine: SyncEngine):
    instances = [
        NestedLevel1(field=1, next_=NestedLevel2(field=2, next_=NestedLevel3(field=3))),
        NestedLevel1(field=4, next_=NestedLevel2(field=5, next_=NestedLevel3(field=6))),
    ]
    sync_engine.save_all(instances)

    fetched = list(sync_engine.find(NestedLevel1))
    assert len(fetched) == 2
    assert fetched[0] in instances
    assert fetched[1] in instances


async def test_reference_with_key_name(aio_engine: AIOEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference(key_name="fancy_key_name")

    instance = M(r=R(field=3))
    assert "fancy_key_name" in instance.model_dump_doc()
    await aio_engine.save(instance)

    fetched = await aio_engine.find_one(M)
    assert fetched is not None
    assert fetched.r.field == 3


def test_sync_reference_with_key_name(sync_engine: SyncEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference(key_name="fancy_key_name")

    instance = M(r=R(field=3))
    assert "fancy_key_name" in instance.model_dump_doc()
    sync_engine.save(instance)

    fetched = sync_engine.find_one(M)
    assert fetched is not None
    assert fetched.r.field == 3


async def test_reference_not_set_in_database(aio_engine: AIOEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference()

    oid = ObjectId()
    await aio_engine.get_collection(M).insert_one({"_id": oid})
    with pytest.raises(DocumentParsingError) as exc_info:
        await aio_engine.find_one(M)
    assert redact_objectid(str(exc_info.value), oid) == snapshot(
        """\
1 validation error for M
r
  Referenced document not found for foreign key 'r' [type=odmantic::referenced_document_not_found, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


def test_sync_reference_not_set_in_database(sync_engine: SyncEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference()

    oid = ObjectId()
    sync_engine.get_collection(M).insert_one({"_id": oid})
    with pytest.raises(DocumentParsingError) as exc_info:
        sync_engine.find_one(M)
    assert redact_objectid(str(exc_info.value), oid) == snapshot(
        """\
1 validation error for M
r
  Referenced document not found for foreign key 'r' [type=odmantic::referenced_document_not_found, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


async def test_reference_incorect_reference_structure(aio_engine: AIOEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference()

    r = R(field=12)
    r_doc = r.model_dump_doc()
    del r_doc["field"]
    m = M(r=r)
    await aio_engine.get_collection(R).insert_one(r_doc)
    await aio_engine.get_collection(M).insert_one(m.model_dump_doc())

    with pytest.raises(DocumentParsingError) as exc_info:
        await aio_engine.find_one(M)
    assert redact_objectid(str(exc_info.value), r.id) == snapshot(
        """\
1 validation error for M
r.field
  Key 'field' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )


def test_sync_reference_incorect_reference_structure(sync_engine: SyncEngine):
    class R(Model):
        field: int

    class M(Model):
        r: R = Reference()

    r = R(field=12)
    r_doc = r.model_dump_doc()
    del r_doc["field"]
    m = M(r=r)
    sync_engine.get_collection(R).insert_one(r_doc)
    sync_engine.get_collection(M).insert_one(m.model_dump_doc())

    with pytest.raises(DocumentParsingError) as exc_info:
        sync_engine.find_one(M)
    assert redact_objectid(str(exc_info.value), r.id) == snapshot(
        """\
1 validation error for M
r.field
  Key 'field' not found in document [type=odmantic::key_not_found_in_document, input_value={'_id': ObjectId('<ObjectId>')}, input_type=dict]\
"""  # noqa: E501
    )



from odmantic import Model, query


class Tree(Model):
    name: str
    average_size: float


Tree.average_size > 2
#> QueryExpression({'average_size': {'$gt': 2}})
Tree.average_size.gt(2)
#> QueryExpression({'average_size': {'$gt': 2}})
query.gt(Tree.average_size, 2)
#> QueryExpression({'average_size': {'$gt': 2}})

Tree.average_size >= 2
#> QueryExpression({'average_size': {'$gte': 2}})
Tree.average_size.gte(2)
#> QueryExpression({'average_size': {'$gte': 2}})
query.gte(Tree.average_size, 2)
#> QueryExpression({'average_size': {'$gte': 2}})



import re
from typing import cast

import pytest

from odmantic import Model
from odmantic.engine import AIOEngine, SyncEngine
from odmantic.query import (
    QueryExpression,
    and_,
    eq,
    gt,
    gte,
    in_,
    lt,
    lte,
    match,
    ne,
    nor_,
    not_in,
    or_,
)

from ..zoo.person import PersonModel

pytestmark = pytest.mark.asyncio


@pytest.fixture(scope="function")
async def person_persisted(aio_engine: AIOEngine):
    initial_instances = [
        PersonModel(first_name="Jean-Pierre", last_name="Pernaud"),
        PersonModel(first_name="Jean-Pierre", last_name="Castaldi"),
        PersonModel(first_name="Michel", last_name="Drucker"),
    ]
    return await aio_engine.save_all(initial_instances)


@pytest.mark.usefixtures("person_persisted")
async def test_and(aio_engine: AIOEngine):
    query = (PersonModel.first_name == "Michel") & (PersonModel.last_name == "Drucker")
    assert query == and_(
        PersonModel.first_name == "Michel", PersonModel.last_name == "Drucker"
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_and(sync_engine: SyncEngine):
    query = (PersonModel.first_name == "Michel") & (PersonModel.last_name == "Drucker")
    assert query == and_(
        PersonModel.first_name == "Michel", PersonModel.last_name == "Drucker"
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
async def test_or(aio_engine: AIOEngine):
    query = (PersonModel.first_name == "Michel") | (PersonModel.last_name == "Castaldi")
    assert query == or_(
        PersonModel.first_name == "Michel", PersonModel.last_name == "Castaldi"
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_or(sync_engine: SyncEngine):
    query = (PersonModel.first_name == "Michel") | (PersonModel.last_name == "Castaldi")
    assert query == or_(
        PersonModel.first_name == "Michel", PersonModel.last_name == "Castaldi"
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
async def test_nor(aio_engine: AIOEngine):
    count = await aio_engine.count(
        PersonModel,
        nor_(PersonModel.first_name == "Michel", PersonModel.last_name == "Castaldi"),
    )
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
async def test_eq(aio_engine: AIOEngine):
    query = cast(QueryExpression, PersonModel.first_name == "Michel")
    assert query == eq(PersonModel.first_name, "Michel")
    count = await aio_engine.count(PersonModel, query)
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
def test_sync_eq(sync_engine: SyncEngine):
    query = cast(QueryExpression, PersonModel.first_name == "Michel")
    assert query == eq(PersonModel.first_name, "Michel")
    count = sync_engine.count(PersonModel, query)
    assert count == 1


@pytest.mark.usefixtures("person_persisted")
async def test_ne(aio_engine: AIOEngine):
    query = PersonModel.first_name != "Michel"
    assert query == ne(PersonModel.first_name, "Michel")
    count = await aio_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_ne(sync_engine: SyncEngine):
    query = PersonModel.first_name != "Michel"
    assert query == ne(PersonModel.first_name, "Michel")
    count = sync_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
async def test_in_(aio_engine: AIOEngine):
    query = in_(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.in_(  # type: ignore
        ["Michel", "Jean-Pierre"]
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 3


@pytest.mark.usefixtures("person_persisted")
def test_sync_in_(sync_engine: SyncEngine):
    query = in_(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.in_(  # type: ignore
        ["Michel", "Jean-Pierre"]
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 3


@pytest.mark.usefixtures("person_persisted")
async def test_in__generator(aio_engine: AIOEngine):
    query = in_(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.in_(  # type: ignore
        e for e in ["Michel", "Jean-Pierre"]
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 3


@pytest.mark.usefixtures("person_persisted")
def test_sync_in__generator(sync_engine: SyncEngine):
    query = in_(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.in_(  # type: ignore
        e for e in ["Michel", "Jean-Pierre"]
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 3


@pytest.mark.usefixtures("person_persisted")
async def test_not_in(aio_engine: AIOEngine):
    query = not_in(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.not_in(  # type: ignore
        ["Michel", "Jean-Pierre"]
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 0


@pytest.mark.usefixtures("person_persisted")
def test_sync_not_in(sync_engine: SyncEngine):
    query = not_in(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.not_in(  # type: ignore
        ["Michel", "Jean-Pierre"]
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 0


@pytest.mark.usefixtures("person_persisted")
async def test_not_in_generator(aio_engine: AIOEngine):
    query = not_in(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.not_in(  # type: ignore
        e for e in ["Michel", "Jean-Pierre"]
    )
    count = await aio_engine.count(PersonModel, query)
    assert count == 0


@pytest.mark.usefixtures("person_persisted")
def test_sync_not_in_generator(sync_engine: SyncEngine):
    query = not_in(PersonModel.first_name, ["Michel", "Jean-Pierre"])
    # TODO allow this with a mypy plugin
    assert query == PersonModel.first_name.not_in(  # type: ignore
        e for e in ["Michel", "Jean-Pierre"]
    )
    count = sync_engine.count(PersonModel, query)
    assert count == 0


class AgedPerson(Model):
    name: str
    age: int


@pytest.fixture(scope="function")
async def aged_person_persisted(aio_engine: AIOEngine):
    initial_instances = [
        AgedPerson(name="Jean-Pierre", age=25),
        AgedPerson(name="Jean-Paul", age=40),
        AgedPerson(name="Michel", age=70),
    ]
    return await aio_engine.save_all(initial_instances)


@pytest.mark.usefixtures("aged_person_persisted")
async def test_gt(aio_engine: AIOEngine):
    query = AgedPerson.age > 40
    assert query == AgedPerson.age.gt(40)  # type: ignore
    assert query == gt(AgedPerson.age, 40)
    count = await aio_engine.count(AgedPerson, query)
    assert count == 1


@pytest.mark.usefixtures("aged_person_persisted")
def test_sync_gt(sync_engine: SyncEngine):
    query = AgedPerson.age > 40
    assert query == AgedPerson.age.gt(40)  # type: ignore
    assert query == gt(AgedPerson.age, 40)
    count = sync_engine.count(AgedPerson, query)
    assert count == 1


@pytest.mark.usefixtures("aged_person_persisted")
async def test_gte(aio_engine: AIOEngine):
    query = AgedPerson.age >= 40
    assert query == AgedPerson.age.gte(40)  # type: ignore
    assert query == gte(AgedPerson.age, 40)
    count = await aio_engine.count(AgedPerson, query)
    assert count == 2


@pytest.mark.usefixtures("aged_person_persisted")
async def test_lt(aio_engine: AIOEngine):
    query = AgedPerson.age < 40
    assert query == AgedPerson.age.lt(40)  # type: ignore
    assert query == lt(AgedPerson.age, 40)
    count = await aio_engine.count(AgedPerson, query)
    assert count == 1


@pytest.mark.usefixtures("aged_person_persisted")
def test_sync_lt(sync_engine: SyncEngine):
    query = AgedPerson.age < 40
    assert query == AgedPerson.age.lt(40)  # type: ignore
    assert query == lt(AgedPerson.age, 40)
    count = sync_engine.count(AgedPerson, query)
    assert count == 1


@pytest.mark.usefixtures("aged_person_persisted")
async def test_lte(aio_engine: AIOEngine):
    query = AgedPerson.age <= 40
    assert query == AgedPerson.age.lte(40)  # type: ignore
    assert query == lte(AgedPerson.age, 40)
    count = await aio_engine.count(AgedPerson, query)
    assert count == 2


@pytest.mark.usefixtures("aged_person_persisted")
def test_sync_lte(sync_engine: SyncEngine):
    query = AgedPerson.age <= 40
    assert query == AgedPerson.age.lte(40)  # type: ignore
    assert query == lte(AgedPerson.age, 40)
    count = sync_engine.count(AgedPerson, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
async def test_match_pattern_string(aio_engine: AIOEngine):
    # TODO allow this with a mypy plugin
    query = PersonModel.first_name.match(r"^Jean-.*")  # type: ignore
    assert query == match(PersonModel.first_name, "^Jean-.*")
    count = await aio_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_match_pattern_string(sync_engine: SyncEngine):
    # TODO allow this with a mypy plugin
    query = PersonModel.first_name.match(r"^Jean-.*")  # type: ignore
    assert query == match(PersonModel.first_name, "^Jean-.*")
    count = sync_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
async def test_match_pattern_compiled(aio_engine: AIOEngine):
    # TODO allow this with a mypy plugin
    r = re.compile(r"^Jean-.*")
    query = PersonModel.first_name.match(r)  # type: ignore
    assert query == match(PersonModel.first_name, r)
    count = await aio_engine.count(PersonModel, query)
    assert count == 2


@pytest.mark.usefixtures("person_persisted")
def test_sync_match_pattern_compiled(sync_engine: SyncEngine):
    # TODO allow this with a mypy plugin
    r = re.compile(r"^Jean-.*")
    query = PersonModel.first_name.match(r)  # type: ignore
    assert query == match(PersonModel.first_name, r)
    count = sync_engine.count(PersonModel, query)
    assert count == 2



from odmantic import SyncEngine, Model


class Player(Model):
    name: str
    game: str


engine = SyncEngine()

leeroy = Player(name="Leeroy Jenkins", game="World of Warcraft")
engine.save(leeroy)

players = [
    Player(name="Shroud", game="Counter-Strike"),
    Player(name="Serral", game="Starcraft"),
    Player(name="TLO", game="Starcraft"),
]
engine.save_all(players)



from typing import ClassVar

from pydantic import ValidationError, model_validator

from odmantic import Model


class SmallRectangle(Model):
    MAX_AREA: ClassVar[float] = 9

    length: float
    width: float

    @model_validator(mode="before")
    def check_width_length(cls, values):
        length = values.get("length", 0)
        width = values.get("width", 0)
        if width > length:
            raise ValueError("width can't be greater than length")
        return values

    @model_validator(mode="before")
    def check_area(cls, values):
        length = values.get("length", 0)
        width = values.get("width", 0)
        if length * width > cls.MAX_AREA:
            raise ValueError(f"area is greater than {cls.MAX_AREA}")
        return values


print(SmallRectangle(length=2, width=1))
# > id=ObjectId('5f81e3c073103f509f97e374'), length=2.0, width=1.0

try:
    SmallRectangle(length=2)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
    width
      field required (type=value_error.missing)
    """

try:
    SmallRectangle(length=2, width=3)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
      Value error, width can't be greater than length
    """

try:
    SmallRectangle(length=4, width=3)
except ValidationError as e:
    print(e)
    """
    1 validation error for SmallRectangle
    __root__
      Value error, area is greater than 9
    """
